import sys
import inspect

from abc import ABCMeta
from logging import Logger

from typing import Callable, Any, Optional, Union, List, Tuple


class DbApiTemplate(object):
    __metaclass__ = ABCMeta

    def __init__(self, conn, logger=None):
        # type: (Any, Logger) -> None
        self.conn = conn
        self.logger = logger
        self.version = sys.version_info[0]

    def get_scalar(self, sql):
        def inner_func(cursor):
            # type: (Any) -> int
            return cursor.execute(sql).fetchone()[0]

        return self.execute_with_cursor(inner_func)

    def one(self, sql, params=(), mapper=None):
        def default_mapper(row):
            return row[0]

        used_mapper = default_mapper
        if mapper is not None:
            used_mapper = mapper

        return self.execute(sql, params, used_mapper)[0]

    def first(self, sql, params=(), mapper=None):
        def default_mapper(row):
            return row[0]

        used_mapper = default_mapper
        if mapper is not None:
            used_mapper = mapper

        ret = self.execute(sql, params, used_mapper)
        if len(ret) > 0:
            return ret[0]
        else:
            return None

    def execute(self, sql, params=(), mapper=None):
        # type: (str, tuple, Optional[Callable[[tuple], Any]]) -> Union[int, list]
        if mapper is None:
            return self.execute_scalar(sql, params)
        else:
            return self.execute_query(sql, params, mapper)

    def execute_query(self, sql, params, mapper):
        # type: (str, tuple, Optional[Callable[[tuple], Any]]) -> Union[int, list]
        def inner_func(cursor):
            # type: (tuple) -> list
            ret = []
            self.debug("Executing query: %s\nParameters: %s", sql, params)
            cursor.execute(sql, params)

            if sys.version_info[0] >= 3:
                has_column_attributes = len(inspect.signature(mapper).parameters) >= 2
            else:
                has_column_attributes = len(inspect.getargspec(mapper)[0]) >= 2

            column_attributes = {}
            for idx in range(len(cursor.description)):
                column_attributes[idx] = cursor.description[idx]

            for row in cursor.fetchall():
                named_tuple = NamedTuple(row, column_attributes)
                if has_column_attributes:
                    value = mapper(named_tuple, column_attributes)
                else:
                    value = mapper(named_tuple)
                ret.append(value)
            return ret

        return self.execute_with_cursor(inner_func)

    def execute_scalar(self, sql, params):
        # type: (str, tuple) -> int
        def inner_func(cursor):
            # type: (Any) -> Tuple[int, Any]
            self.debug("Executing query: %s\nParameters:%s", sql, params)
            cursor.execute(sql, params)
            self.debug("Query executed")
            row_count = cursor.rowcount
            ret_data = None
            try:
                row = cursor.fetchone()
                ret_data = row[0]
            except:
                pass
            return row_count, ret_data

        return self.execute_with_cursor(inner_func)

    def execute_with_cursor(self, method):
        # type: (Callable[[tuple], Any]) -> Any
        cursor = None
        try:
            cursor = self.conn.cursor()
            return method(cursor)
        finally:
            if cursor is not None:
                cursor.close()

    def execute_many(self, sql, params):
        # type: (str, List[Tuple]) -> None
        self.conn.executemany(sql, params)

    def debug(self, msg, *params):
        if self.logger is None:
            return
        self.logger.debug(msg, params)


class UnitOfWork(object):
    def __init__(self, conn):
        # type: (Any) -> None
        self.conn = conn

    def with_transaction(self, method):
        try:
            ret = method()
            self.conn.commit()
            return ret
        except:
            self.conn.rollback()
            raise


class NamedTuple(object):
    def __init__(self, row, column_attributes):
        super(NamedTuple, self).__init__()
        self.row = row
        self.column_attributes = column_attributes
        self.col_map = {}
        for i in range(len(column_attributes)):
            self.col_map[column_attributes[i][0]] = i

    def __len__(self):
        return len(self.row)

    def __getitem__(self, item):
        index = item
        if isinstance(item, str):
            index = self.col_map[item]
        return self.row[index]

    def __str__(self):
        if len(self.row) == 0:
            return "()"
        ret = "("
        for el in self.row:
            ret += repr(el) + ", "
        ret = ret[0:-2]
        ret += ")"
        attrs = ""
        for i in range(len(self.row)):
            attrs += self.column_attributes[i][0] + ", "
        attrs = attrs[0:-1]
        return ret + " - " + attrs



from user_levels import UserLevels
from GLAccount import GLAccount
from TransferTypeEnum import TransferType
from sale_types import SaleTypes
from navigation import Navigation
from navigation_product import NavigationProduct
from ._CloseOperator import CloseOperator
from ._PaymentTypeEnum import PaymentType
from ._Course import Course
from ._AuthorizationResponse import AuthorizationResponse



class AuditEvent(object):
    PAID = "PAID"
    TABLE_CLOSED = "TABLE_CLOSED"
    VOID_ORDER = "VOID_ORDER"
    OPERATOR_LOGOUT = "OPERATORLOGOUT"
    PRODUCTION_ORDER_TIMES = "PRODUCTION_ORDER_TIMES"
    ORDER_REPROCESS = "ORDER_REPROCESS"
    CLOSE_STORE_DAY = "CLOSE_STORE_DAY"
    OPEN_STORE_DAY = "OPEN_STORE_DAY"



from functools import update_wrapper
from io import BytesIO

from .._compat import to_native
from .._compat import to_unicode
from .._compat import wsgi_decoding_dance
from .._compat import wsgi_get_bytes
from ..datastructures import CombinedMultiDict
from ..datastructures import EnvironHeaders
from ..datastructures import ImmutableList
from ..datastructures import ImmutableMultiDict
from ..datastructures import iter_multi_items
from ..datastructures import MultiDict
from ..formparser import default_stream_factory
from ..formparser import FormDataParser
from ..http import parse_cookie
from ..http import parse_list_header
from ..http import parse_options_header
from ..urls import url_decode
from ..utils import cached_property
from ..utils import environ_property
from ..wsgi import get_content_length
from ..wsgi import get_current_url
from ..wsgi import get_host
from ..wsgi import get_input_stream


class BaseRequest(object):
    """Very basic request object.  This does not implement advanced stuff like
    entity tag parsing or cache controls.  The request object is created with
    the WSGI environment as first argument and will add itself to the WSGI
    environment as ``'werkzeug.request'`` unless it's created with
    `populate_request` set to False.

    There are a couple of mixins available that add additional functionality
    to the request object, there is also a class called `Request` which
    subclasses `BaseRequest` and all the important mixins.

    It's a good idea to create a custom subclass of the :class:`BaseRequest`
    and add missing functionality either via mixins or direct implementation.
    Here an example for such subclasses::

        from werkzeug.wrappers import BaseRequest, ETagRequestMixin

        class Request(BaseRequest, ETagRequestMixin):
            pass

    Request objects are **read only**.  As of 0.5 modifications are not
    allowed in any place.  Unlike the lower level parsing functions the
    request object will use immutable objects everywhere possible.

    Per default the request object will assume all the text data is `utf-8`
    encoded.  Please refer to :doc:`the unicode chapter </unicode>` for more
    details about customizing the behavior.

    Per default the request object will be added to the WSGI
    environment as `werkzeug.request` to support the debugging system.
    If you don't want that, set `populate_request` to `False`.

    If `shallow` is `True` the environment is initialized as shallow
    object around the environ.  Every operation that would modify the
    environ in any way (such as consuming form data) raises an exception
    unless the `shallow` attribute is explicitly set to `False`.  This
    is useful for middlewares where you don't want to consume the form
    data by accident.  A shallow request is not populated to the WSGI
    environment.

    .. versionchanged:: 0.5
       read-only mode was enforced by using immutables classes for all
       data.
    """

    #: the charset for the request, defaults to utf-8
    charset = "utf-8"

    #: the error handling procedure for errors, defaults to 'replace'
    encoding_errors = "replace"

    #: the maximum content length.  This is forwarded to the form data
    #: parsing function (:func:`parse_form_data`).  When set and the
    #: :attr:`form` or :attr:`files` attribute is accessed and the
    #: parsing fails because more than the specified value is transmitted
    #: a :exc:`~werkzeug.exceptions.RequestEntityTooLarge` exception is raised.
    #:
    #: Have a look at :ref:`dealing-with-request-data` for more details.
    #:
    #: .. versionadded:: 0.5
    max_content_length = None

    #: the maximum form field size.  This is forwarded to the form data
    #: parsing function (:func:`parse_form_data`).  When set and the
    #: :attr:`form` or :attr:`files` attribute is accessed and the
    #: data in memory for post data is longer than the specified value a
    #: :exc:`~werkzeug.exceptions.RequestEntityTooLarge` exception is raised.
    #:
    #: Have a look at :ref:`dealing-with-request-data` for more details.
    #:
    #: .. versionadded:: 0.5
    max_form_memory_size = None

    #: the class to use for `args` and `form`.  The default is an
    #: :class:`~werkzeug.datastructures.ImmutableMultiDict` which supports
    #: multiple values per key.  alternatively it makes sense to use an
    #: :class:`~werkzeug.datastructures.ImmutableOrderedMultiDict` which
    #: preserves order or a :class:`~werkzeug.datastructures.ImmutableDict`
    #: which is the fastest but only remembers the last key.  It is also
    #: possible to use mutable structures, but this is not recommended.
    #:
    #: .. versionadded:: 0.6
    parameter_storage_class = ImmutableMultiDict

    #: the type to be used for list values from the incoming WSGI environment.
    #: By default an :class:`~werkzeug.datastructures.ImmutableList` is used
    #: (for example for :attr:`access_list`).
    #:
    #: .. versionadded:: 0.6
    list_storage_class = ImmutableList

    #: The type to be used for dict values from the incoming WSGI
    #: environment. (For example for :attr:`cookies`.) By default an
    #: :class:`~werkzeug.datastructures.ImmutableMultiDict` is used.
    #:
    #: .. versionchanged:: 1.0.0
    #:     Changed to ``ImmutableMultiDict`` to support multiple values.
    #:
    #: .. versionadded:: 0.6
    dict_storage_class = ImmutableMultiDict

    #: The form data parser that shoud be used.  Can be replaced to customize
    #: the form date parsing.
    form_data_parser_class = FormDataParser

    #: Optionally a list of hosts that is trusted by this request.  By default
    #: all hosts are trusted which means that whatever the client sends the
    #: host is will be accepted.
    #:
    #: Because `Host` and `X-Forwarded-Host` headers can be set to any value by
    #: a malicious client, it is recommended to either set this property or
    #: implement similar validation in the proxy (if application is being run
    #: behind one).
    #:
    #: .. versionadded:: 0.9
    trusted_hosts = None

    #: Indicates whether the data descriptor should be allowed to read and
    #: buffer up the input stream.  By default it's enabled.
    #:
    #: .. versionadded:: 0.9
    disable_data_descriptor = False

    def __init__(self, environ, populate_request=True, shallow=False):
        self.environ = environ
        if populate_request and not shallow:
            self.environ["werkzeug.request"] = self
        self.shallow = shallow

    def __repr__(self):
        # make sure the __repr__ even works if the request was created
        # from an invalid WSGI environment.  If we display the request
        # in a debug session we don't want the repr to blow up.
        args = []
        try:
            args.append("'%s'" % to_native(self.url, self.url_charset))
            args.append("[%s]" % self.method)
        except Exception:
            args.append("(invalid WSGI environ)")

        return "<%s %s>" % (self.__class__.__name__, " ".join(args))

    @property
    def url_charset(self):
        """The charset that is assumed for URLs.  Defaults to the value
        of :attr:`charset`.

        .. versionadded:: 0.6
        """
        return self.charset

    @classmethod
    def from_values(cls, *args, **kwargs):
        """Create a new request object based on the values provided.  If
        environ is given missing values are filled from there.  This method is
        useful for small scripts when you need to simulate a request from an URL.
        Do not use this method for unittesting, there is a full featured client
        object (:class:`Client`) that allows to create multipart requests,
        support for cookies etc.

        This accepts the same options as the
        :class:`~werkzeug.test.EnvironBuilder`.

        .. versionchanged:: 0.5
           This method now accepts the same arguments as
           :class:`~werkzeug.test.EnvironBuilder`.  Because of this the
           `environ` parameter is now called `environ_overrides`.

        :return: request object
        """
        from ..test import EnvironBuilder

        charset = kwargs.pop("charset", cls.charset)
        kwargs["charset"] = charset
        builder = EnvironBuilder(*args, **kwargs)
        try:
            return builder.get_request(cls)
        finally:
            builder.close()

    @classmethod
    def application(cls, f):
        """Decorate a function as responder that accepts the request as
        the last argument.  This works like the :func:`responder`
        decorator but the function is passed the request object as the
        last argument and the request object will be closed
        automatically::

            @Request.application
            def my_wsgi_app(request):
                return Response('Hello World!')

        As of Werkzeug 0.14 HTTP exceptions are automatically caught and
        converted to responses instead of failing.

        :param f: the WSGI callable to decorate
        :return: a new WSGI callable
        """
        #: return a callable that wraps the -2nd argument with the request
        #: and calls the function with all the arguments up to that one and
        #: the request.  The return value is then called with the latest
        #: two arguments.  This makes it possible to use this decorator for
        #: both standalone WSGI functions as well as bound methods and
        #: partially applied functions.
        from ..exceptions import HTTPException

        def application(*args):
            request = cls(args[-2])
            with request:
                try:
                    resp = f(*args[:-2] + (request,))
                except HTTPException as e:
                    resp = e.get_response(args[-2])
                return resp(*args[-2:])

        return update_wrapper(application, f)

    def _get_file_stream(
        self, total_content_length, content_type, filename=None, content_length=None
    ):
        """Called to get a stream for the file upload.

        This must provide a file-like class with `read()`, `readline()`
        and `seek()` methods that is both writeable and readable.

        The default implementation returns a temporary file if the total
        content length is higher than 500KB.  Because many browsers do not
        provide a content length for the files only the total content
        length matters.

        :param total_content_length: the total content length of all the
                                     data in the request combined.  This value
                                     is guaranteed to be there.
        :param content_type: the mimetype of the uploaded file.
        :param filename: the filename of the uploaded file.  May be `None`.
        :param content_length: the length of this file.  This value is usually
                               not provided because webbrowsers do not provide
                               this value.
        """
        return default_stream_factory(
            total_content_length=total_content_length,
            filename=filename,
            content_type=content_type,
            content_length=content_length,
        )

    @property
    def want_form_data_parsed(self):
        """Returns True if the request method carries content.  As of
        Werkzeug 0.9 this will be the case if a content type is transmitted.

        .. versionadded:: 0.8
        """
        return bool(self.environ.get("CONTENT_TYPE"))

    def make_form_data_parser(self):
        """Creates the form data parser. Instantiates the
        :attr:`form_data_parser_class` with some parameters.

        .. versionadded:: 0.8
        """
        return self.form_data_parser_class(
            self._get_file_stream,
            self.charset,
            self.encoding_errors,
            self.max_form_memory_size,
            self.max_content_length,
            self.parameter_storage_class,
        )

    def _load_form_data(self):
        """Method used internally to retrieve submitted data.  After calling
        this sets `form` and `files` on the request object to multi dicts
        filled with the incoming form data.  As a matter of fact the input
        stream will be empty afterwards.  You can also call this method to
        force the parsing of the form data.

        .. versionadded:: 0.8
        """
        # abort early if we have already consumed the stream
        if "form" in self.__dict__:
            return

        _assert_not_shallow(self)

        if self.want_form_data_parsed:
            content_type = self.environ.get("CONTENT_TYPE", "")
            content_length = get_content_length(self.environ)
            mimetype, options = parse_options_header(content_type)
            parser = self.make_form_data_parser()
            data = parser.parse(
                self._get_stream_for_parsing(), mimetype, content_length, options
            )
        else:
            data = (
                self.stream,
                self.parameter_storage_class(),
                self.parameter_storage_class(),
            )

        # inject the values into the instance dict so that we bypass
        # our cached_property non-data descriptor.
        d = self.__dict__
        d["stream"], d["form"], d["files"] = data

    def _get_stream_for_parsing(self):
        """This is the same as accessing :attr:`stream` with the difference
        that if it finds cached data from calling :meth:`get_data` first it
        will create a new stream out of the cached data.

        .. versionadded:: 0.9.3
        """
        cached_data = getattr(self, "_cached_data", None)
        if cached_data is not None:
            return BytesIO(cached_data)
        return self.stream

    def close(self):
        """Closes associated resources of this request object.  This
        closes all file handles explicitly.  You can also use the request
        object in a with statement which will automatically close it.

        .. versionadded:: 0.9
        """
        files = self.__dict__.get("files")
        for _key, value in iter_multi_items(files or ()):
            value.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    @cached_property
    def stream(self):
        """
        If the incoming form data was not encoded with a known mimetype
        the data is stored unmodified in this stream for consumption.  Most
        of the time it is a better idea to use :attr:`data` which will give
        you that data as a string.  The stream only returns the data once.

        Unlike :attr:`input_stream` this stream is properly guarded that you
        can't accidentally read past the length of the input.  Werkzeug will
        internally always refer to this stream to read data which makes it
        possible to wrap this object with a stream that does filtering.

        .. versionchanged:: 0.9
           This stream is now always available but might be consumed by the
           form parser later on.  Previously the stream was only set if no
           parsing happened.
        """
        _assert_not_shallow(self)
        return get_input_stream(self.environ)

    input_stream = environ_property(
        "wsgi.input",
        """The WSGI input stream.

        In general it's a bad idea to use this one because you can
        easily read past the boundary.  Use the :attr:`stream`
        instead.""",
    )

    @cached_property
    def args(self):
        """The parsed URL parameters (the part in the URL after the question
        mark).

        By default an
        :class:`~werkzeug.datastructures.ImmutableMultiDict`
        is returned from this function.  This can be changed by setting
        :attr:`parameter_storage_class` to a different type.  This might
        be necessary if the order of the form data is important.
        """
        return url_decode(
            wsgi_get_bytes(self.environ.get("QUERY_STRING", "")),
            self.url_charset,
            errors=self.encoding_errors,
            cls=self.parameter_storage_class,
        )

    @cached_property
    def data(self):
        """
        Contains the incoming request data as string in case it came with
        a mimetype Werkzeug does not handle.
        """

        if self.disable_data_descriptor:
            raise AttributeError("data descriptor is disabled")
        # XXX: this should eventually be deprecated.

        # We trigger form data parsing first which means that the descriptor
        # will not cache the data that would otherwise be .form or .files
        # data.  This restores the behavior that was there in Werkzeug
        # before 0.9.  New code should use :meth:`get_data` explicitly as
        # this will make behavior explicit.
        return self.get_data(parse_form_data=True)

    def get_data(self, cache=True, as_text=False, parse_form_data=False):
        """This reads the buffered incoming data from the client into one
        bytestring.  By default this is cached but that behavior can be
        changed by setting `cache` to `False`.

        Usually it's a bad idea to call this method without checking the
        content length first as a client could send dozens of megabytes or more
        to cause memory problems on the server.

        Note that if the form data was already parsed this method will not
        return anything as form data parsing does not cache the data like
        this method does.  To implicitly invoke form data parsing function
        set `parse_form_data` to `True`.  When this is done the return value
        of this method will be an empty string if the form parser handles
        the data.  This generally is not necessary as if the whole data is
        cached (which is the default) the form parser will used the cached
        data to parse the form data.  Please be generally aware of checking
        the content length first in any case before calling this method
        to avoid exhausting server memory.

        If `as_text` is set to `True` the return value will be a decoded
        unicode string.

        .. versionadded:: 0.9
        """
        rv = getattr(self, "_cached_data", None)
        if rv is None:
            if parse_form_data:
                self._load_form_data()
            rv = self.stream.read()
            if cache:
                self._cached_data = rv
        if as_text:
            rv = rv.decode(self.charset, self.encoding_errors)
        return rv

    @cached_property
    def form(self):
        """The form parameters.  By default an
        :class:`~werkzeug.datastructures.ImmutableMultiDict`
        is returned from this function.  This can be changed by setting
        :attr:`parameter_storage_class` to a different type.  This might
        be necessary if the order of the form data is important.

        Please keep in mind that file uploads will not end up here, but instead
        in the :attr:`files` attribute.

        .. versionchanged:: 0.9

            Previous to Werkzeug 0.9 this would only contain form data for POST
            and PUT requests.
        """
        self._load_form_data()
        return self.form

    @cached_property
    def values(self):
        """A :class:`werkzeug.datastructures.CombinedMultiDict` that combines
        :attr:`args` and :attr:`form`."""
        args = []
        for d in self.args, self.form:
            if not isinstance(d, MultiDict):
                d = MultiDict(d)
            args.append(d)
        return CombinedMultiDict(args)

    @cached_property
    def files(self):
        """:class:`~werkzeug.datastructures.MultiDict` object containing
        all uploaded files.  Each key in :attr:`files` is the name from the
        ``<input type="file" name="">``.  Each value in :attr:`files` is a
        Werkzeug :class:`~werkzeug.datastructures.FileStorage` object.

        It basically behaves like a standard file object you know from Python,
        with the difference that it also has a
        :meth:`~werkzeug.datastructures.FileStorage.save` function that can
        store the file on the filesystem.

        Note that :attr:`files` will only contain data if the request method was
        POST, PUT or PATCH and the ``<form>`` that posted to the request had
        ``enctype="multipart/form-data"``.  It will be empty otherwise.

        See the :class:`~werkzeug.datastructures.MultiDict` /
        :class:`~werkzeug.datastructures.FileStorage` documentation for
        more details about the used data structure.
        """
        self._load_form_data()
        return self.files

    @cached_property
    def cookies(self):
        """A :class:`dict` with the contents of all cookies transmitted with
        the request."""
        return parse_cookie(
            self.environ,
            self.charset,
            self.encoding_errors,
            cls=self.dict_storage_class,
        )

    @cached_property
    def headers(self):
        """The headers from the WSGI environ as immutable
        :class:`~werkzeug.datastructures.EnvironHeaders`.
        """
        return EnvironHeaders(self.environ)

    @cached_property
    def path(self):
        """Requested path as unicode.  This works a bit like the regular path
        info in the WSGI environment but will always include a leading slash,
        even if the URL root is accessed.
        """
        raw_path = wsgi_decoding_dance(
            self.environ.get("PATH_INFO") or "", self.charset, self.encoding_errors
        )
        return "/" + raw_path.lstrip("/")

    @cached_property
    def full_path(self):
        """Requested path as unicode, including the query string."""
        return self.path + u"?" + to_unicode(self.query_string, self.url_charset)

    @cached_property
    def script_root(self):
        """The root path of the script without the trailing slash."""
        raw_path = wsgi_decoding_dance(
            self.environ.get("SCRIPT_NAME") or "", self.charset, self.encoding_errors
        )
        return raw_path.rstrip("/")

    @cached_property
    def url(self):
        """The reconstructed current URL as IRI.
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(self.environ, trusted_hosts=self.trusted_hosts)

    @cached_property
    def base_url(self):
        """Like :attr:`url` but without the querystring
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(
            self.environ, strip_querystring=True, trusted_hosts=self.trusted_hosts
        )

    @cached_property
    def url_root(self):
        """The full URL root (with hostname), this is the application
        root as IRI.
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(self.environ, True, trusted_hosts=self.trusted_hosts)

    @cached_property
    def host_url(self):
        """Just the host with scheme as IRI.
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(
            self.environ, host_only=True, trusted_hosts=self.trusted_hosts
        )

    @cached_property
    def host(self):
        """Just the host including the port if available.
        See also: :attr:`trusted_hosts`.
        """
        return get_host(self.environ, trusted_hosts=self.trusted_hosts)

    query_string = environ_property(
        "QUERY_STRING",
        "",
        read_only=True,
        load_func=wsgi_get_bytes,
        doc="The URL parameters as raw bytestring.",
    )
    method = environ_property(
        "REQUEST_METHOD",
        "GET",
        read_only=True,
        load_func=lambda x: x.upper(),
        doc="The request method. (For example ``'GET'`` or ``'POST'``).",
    )

    @cached_property
    def access_route(self):
        """If a forwarded header exists this is a list of all ip addresses
        from the client ip to the last proxy server.
        """
        if "HTTP_X_FORWARDED_FOR" in self.environ:
            return self.list_storage_class(
                parse_list_header(self.environ["HTTP_X_FORWARDED_FOR"])
            )
        elif "REMOTE_ADDR" in self.environ:
            return self.list_storage_class([self.environ["REMOTE_ADDR"]])
        return self.list_storage_class()

    @property
    def remote_addr(self):
        """The remote address of the client."""
        return self.environ.get("REMOTE_ADDR")

    remote_user = environ_property(
        "REMOTE_USER",
        doc="""If the server supports user authentication, and the
        script is protected, this attribute contains the username the
        user has authenticated as.""",
    )
    scheme = environ_property(
        "wsgi.url_scheme",
        doc="""
        URL scheme (http or https).

        .. versionadded:: 0.7""",
    )
    is_secure = property(
        lambda self: self.environ["wsgi.url_scheme"] == "https",
        doc="`True` if the request is secure.",
    )
    is_multithread = environ_property(
        "wsgi.multithread",
        doc="""boolean that is `True` if the application is served by a
        multithreaded WSGI server.""",
    )
    is_multiprocess = environ_property(
        "wsgi.multiprocess",
        doc="""boolean that is `True` if the application is served by a
        WSGI server that spawns multiple processes.""",
    )
    is_run_once = environ_property(
        "wsgi.run_once",
        doc="""boolean that is `True` if the application will be
        executed only once in a process lifetime.  This is the case for
        CGI for example, but it's not guaranteed that the execution only
        happens one time.""",
    )


def _assert_not_shallow(request):
    if request.shallow:
        raise RuntimeError(
            "A shallow request tried to consume form data. If you really"
            " want to do that, set `shallow` to False."
        )



from typing import Optional  # noqa
from datetime import datetime


class DefaultHeaderDto(object):
    def __init__(self,
                 report_type,
                 initial_date,
                 end_date, pos,
                 operator_id,
                 operator_name,
                 store_code,
                 generated_time):
        # type: (str, datetime, datetime, Optional[int], Optional[str], Optional[str], str, datetime) -> None
        self.report_type = report_type
        self.initial_date = initial_date
        self.end_date = end_date
        self.pos = pos
        self.operator_id = operator_id
        self.operator_name = operator_name
        self.store_code = store_code
        self.generated_time = generated_time



# -*- coding: utf-8 -*-
from threading import Thread

import bustoken
from typing import Optional

from actions import mb_context
from sysactions import (
    get_model,
    action,
    show_any_dialog,
)
from systools import sys_log_exception


@action
def show_daily_goals(pos_id, operator_id, show_amount_chart, show_item_chart, show_operator_table, show_average_chart):
    update_daily_goals(pos_id)

    return show_any_dialog(pos_id, "dailyGoals", "message", "title", "", "", "", 600000,
                           operator_id + "|" + show_amount_chart + "|" + show_item_chart + "|" +
                           show_operator_table + "|" + show_average_chart,
                           "", "", "", None, False)


def update_daily_goals(
    pos_id,             # type: str
    synchronous=True,   # type: Optional[bool]
):
    # type: (...) -> None

    if synchronous:
        _send__message_to_update_goals(pos_id)
        return

    Thread(target=_send__message_to_update_goals, args=(pos_id,)).start()


def _send__message_to_update_goals(pos_id):
    try:
        model = get_model(pos_id)
        daily_goals = model.findall("Custom[@name='DAILY_GOALS']")
        if len(daily_goals) == 0:
            mb_context.MB_EasySendMessage("DailyGoals", bustoken.TK_DAILY_GOALS_UPDATE_TOTAL_SOLD)
    except BaseException as _:
        sys_log_exception("[update_daily_goals]")



from typing import List


class TableReportFilterDto(object):
    def __init__(
            self,
            pos_id,     # type: str
            table_id,   # type: str
            orders      # type: List[str]
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.table_id = table_id
        self.orders = orders



# -*- coding: utf-8 -*-
"""upload_docs

Implements a Distutils 'upload_docs' subcommand (upload documentation to
PyPI's pythonhosted.org).
"""

from base64 import standard_b64encode
from distutils import log
from distutils.errors import DistutilsOptionError
import os
import socket
import zipfile
import tempfile
import shutil
import itertools
import functools

from setuptools.extern import six
from setuptools.extern.six.moves import http_client, urllib

from pkg_resources import iter_entry_points
from .upload import upload


def _encode(s):
    errors = 'strict' if six.PY2 else 'surrogateescape'
    return s.encode('utf-8', errors)


class upload_docs(upload):
    # override the default repository as upload_docs isn't
    # supported by Warehouse (and won't be).
    DEFAULT_REPOSITORY = 'https://pypi.python.org/pypi/'

    description = 'Upload documentation to PyPI'

    user_options = [
        ('repository=', 'r',
         "url of repository [default: %s]" % upload.DEFAULT_REPOSITORY),
        ('show-response', None,
         'display full response text from server'),
        ('upload-dir=', None, 'directory to upload'),
    ]
    boolean_options = upload.boolean_options

    def has_sphinx(self):
        if self.upload_dir is None:
            for ep in iter_entry_points('distutils.commands', 'build_sphinx'):
                return True

    sub_commands = [('build_sphinx', has_sphinx)]

    def initialize_options(self):
        upload.initialize_options(self)
        self.upload_dir = None
        self.target_dir = None

    def finalize_options(self):
        upload.finalize_options(self)
        if self.upload_dir is None:
            if self.has_sphinx():
                build_sphinx = self.get_finalized_command('build_sphinx')
                self.target_dir = build_sphinx.builder_target_dir
            else:
                build = self.get_finalized_command('build')
                self.target_dir = os.path.join(build.build_base, 'docs')
        else:
            self.ensure_dirname('upload_dir')
            self.target_dir = self.upload_dir
        if 'pypi.python.org' in self.repository:
            log.warn("Upload_docs command is deprecated. Use RTD instead.")
        self.announce('Using upload directory %s' % self.target_dir)

    def create_zipfile(self, filename):
        zip_file = zipfile.ZipFile(filename, "w")
        try:
            self.mkpath(self.target_dir)  # just in case
            for root, dirs, files in os.walk(self.target_dir):
                if root == self.target_dir and not files:
                    tmpl = "no files found in upload directory '%s'"
                    raise DistutilsOptionError(tmpl % self.target_dir)
                for name in files:
                    full = os.path.join(root, name)
                    relative = root[len(self.target_dir):].lstrip(os.path.sep)
                    dest = os.path.join(relative, name)
                    zip_file.write(full, dest)
        finally:
            zip_file.close()

    def run(self):
        # Run sub commands
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

        tmp_dir = tempfile.mkdtemp()
        name = self.distribution.metadata.get_name()
        zip_file = os.path.join(tmp_dir, "%s.zip" % name)
        try:
            self.create_zipfile(zip_file)
            self.upload_file(zip_file)
        finally:
            shutil.rmtree(tmp_dir)

    @staticmethod
    def _build_part(item, sep_boundary):
        key, values = item
        title = '\nContent-Disposition: form-data; name="%s"' % key
        # handle multiple entries for the same name
        if not isinstance(values, list):
            values = [values]
        for value in values:
            if isinstance(value, tuple):
                title += '; filename="%s"' % value[0]
                value = value[1]
            else:
                value = _encode(value)
            yield sep_boundary
            yield _encode(title)
            yield b"\n\n"
            yield value
            if value and value[-1:] == b'\r':
                yield b'\n'  # write an extra newline (lurve Macs)

    @classmethod
    def _build_multipart(cls, data):
        """
        Build up the MIME payload for the POST data
        """
        boundary = b'--------------GHSKFJDLGDS7543FJKLFHRE75642756743254'
        sep_boundary = b'\n--' + boundary
        end_boundary = sep_boundary + b'--'
        end_items = end_boundary, b"\n",
        builder = functools.partial(
            cls._build_part,
            sep_boundary=sep_boundary,
        )
        part_groups = map(builder, data.items())
        parts = itertools.chain.from_iterable(part_groups)
        body_items = itertools.chain(parts, end_items)
        content_type = 'multipart/form-data; boundary=%s' % boundary.decode('ascii')
        return b''.join(body_items), content_type

    def upload_file(self, filename):
        with open(filename, 'rb') as f:
            content = f.read()
        meta = self.distribution.metadata
        data = {
            ':action': 'doc_upload',
            'name': meta.get_name(),
            'content': (os.path.basename(filename), content),
        }
        # set up the authentication
        credentials = _encode(self.username + ':' + self.password)
        credentials = standard_b64encode(credentials)
        if not six.PY2:
            credentials = credentials.decode('ascii')
        auth = "Basic " + credentials

        body, ct = self._build_multipart(data)

        msg = "Submitting documentation to %s" % (self.repository)
        self.announce(msg, log.INFO)

        # build the Request
        # We can't use urllib2 since we need to send the Basic
        # auth right with the first request
        schema, netloc, url, params, query, fragments = \
            urllib.parse.urlparse(self.repository)
        assert not params and not query and not fragments
        if schema == 'http':
            conn = http_client.HTTPConnection(netloc)
        elif schema == 'https':
            conn = http_client.HTTPSConnection(netloc)
        else:
            raise AssertionError("unsupported schema " + schema)

        data = ''
        try:
            conn.connect()
            conn.putrequest("POST", url)
            content_type = ct
            conn.putheader('Content-type', content_type)
            conn.putheader('Content-length', str(len(body)))
            conn.putheader('Authorization', auth)
            conn.endheaders()
            conn.send(body)
        except socket.error as e:
            self.announce(str(e), log.ERROR)
            return

        r = conn.getresponse()
        if r.status == 200:
            msg = 'Server response (%s): %s' % (r.status, r.reason)
            self.announce(msg, log.INFO)
        elif r.status == 301:
            location = r.getheader('Location')
            if location is None:
                location = 'https://pythonhosted.org/%s/' % meta.get_name()
            msg = 'Upload successful. Visit %s' % location
            self.announce(msg, log.INFO)
        else:
            msg = 'Upload failed (%s): %s' % (r.status, r.reason)
            self.announce(msg, log.ERROR)
        if self.show_response:
            print('-' * 75, r.read(), '-' * 75)



# CER decoder
from pyasn1.type import univ
from pyasn1.codec.ber import decoder
from pyasn1.compat.octets import oct2int
from pyasn1 import error

class BooleanDecoder(decoder.AbstractSimpleDecoder):
    protoComponent = univ.Boolean(0)
    def valueDecoder(self, fullSubstrate, substrate, asn1Spec, tagSet, length,
                     state, decodeFun, substrateFun):
        head, tail = substrate[:length], substrate[length:]
        if not head or length != 1:
            raise error.PyAsn1Error('Not single-octet Boolean payload')
        byte = oct2int(head[0])
        # CER/DER specifies encoding of TRUE as 0xFF and FALSE as 0x0, while
        # BER allows any non-zero value as TRUE; cf. sections 8.2.2. and 11.1 
        # in http://www.itu.int/ITU-T/studygroups/com17/languages/X.690-0207.pdf
        if byte == 0xff:
            value = 1
        elif byte == 0x00:
            value = 0
        else:
            raise error.PyAsn1Error('Unexpected Boolean payload: %s' % byte)
        return self._createComponent(asn1Spec, tagSet, value), tail

tagMap = decoder.tagMap.copy()
tagMap.update({
    univ.Boolean.tagSet: BooleanDecoder()
    })

typeMap = decoder.typeMap

class Decoder(decoder.Decoder): pass

decode = Decoder(tagMap, decoder.typeMap)



import copy
import json
import mw_helper

from xml.etree import cElementTree as eTree
from sysactions import get_model, translate_message, set_custom, show_confirmation
from typing import List, Optional, Dict

from application.interactor import ServicePeriodInteractor, NavigationInteractor
from actions.products import getProducts

DEFAULT_NAME = "{}##null##False"


def get_selection_service_period(pos_id):
    # type: (int) -> Optional[Dict]

    services = ServicePeriodInteractor().get_services_period()
    if len(services) == 0:
        mw_helper.show_message_dialog(pos_id, "$ERROR", "$SERVICES_PERIOD_NOT_FOUND")
        return None

    price_lists = _format_price_list_for_special_catalog(pos_id, services)
    data = "|".join(price_lists)
    index = mw_helper.show_filtered_list_box_dialog(pos_id, data, "$SELECT_SERVICE_PERIOD", "", "NOFILTER")
    if index in (None, ''):
        return None

    if index > 0:
        data = services[index - 1]
        service_period_id = int(data.get("id"))
        service_period = data.get("priceListId", "")
        name = data.get("name", "")
        is_valid = data.get("isValid", False)
        abbreviation = data.get("priceListAbbr", "")
        confirmation = _confirm_service_period_schedule(pos_id, is_valid, name)
        if confirmation:
            return {"servicePeriodId": service_period_id, "servicePeriod": service_period, "servicePeriodName": name,
                    "servicePeriodAbbr": abbreviation}
        else:
            return None

    return {}


def _confirm_service_period_schedule(pos_id, is_valid, name):
    # type: (int, bool, str) -> bool

    confirmation = True
    if not is_valid:
        confirmation = show_confirmation(pos_id, message="$SERVICE_PERIOD_OUT_OF_SCHEDULE_RECORDED|{0}".format(name))
    return confirmation


def set_services_period_all_pos(pos_id, service_period):
    # type: (int, Dict) -> None

    from msgbusboundary import MsgBusTableService as tableService

    model = get_model(pos_id)
    if "servicePeriodName" in service_period:
        service_period_name = service_period["servicePeriodName"]
    else:
        service_period_name = translate_message(model, "NONE")
    sectors = _get_sectors_services_period(pos_id, model, service_period_name)
    if sectors is None:
        return

    special_catalog_sectors = dict(zip(sectors, [copy.deepcopy(service_period)] * len(sectors)))

    pos_list = tableService().get_exclusive_pos_list()
    data = _get_json_data_service_period(service_period, special_catalog_sectors)
    for pos_number in pos_list:
        set_custom(pos_number, 'SERVICE_PERIOD', data, True)

    getProducts(pos_id, force=True)
    # TODO: Update all POS uis


def _get_json_data_service_period(service_period, sectors):
    # type: (Dict, Dict) -> str

    service_period = "{}" if service_period == '' else service_period
    service_period_id = service_period.get("servicePeriodId", None)
    data = {}
    if len(sectors) == 0:
        data = service_period
        data["servicePeriodId"] = service_period_id or ''
        data["servicePeriodName"] = data.get("servicePeriodName", None)
        data["servicePeriodAbbr"] = data.get("servicePeriodAbbr", '')
    nav_interactor = NavigationInteractor()
    data["navigationsRestricted"] = nav_interactor.get_navigations_restricted_to_service_period(service_period_id)
    data["sectors"] = sectors

    return json.dumps(data, default=lambda o: getattr(o, '__dict__', str(o)))


def _format_price_list_for_special_catalog(pos_id, services):
    # type: (int, List) -> List

    services_list = ["{}##{}".format(service.get("name"), service.get("isValid")) for service in services]
    model = get_model(pos_id)
    services_list.insert(0, translate_message(model, 'NONE'))
    return services_list


def _get_sectors_services_period(pos_id, model, service_period_name):
    # type: (int, eTree.XML, str) -> Optional[List]

    from msgbusboundary import MsgBusTableService as tableService

    sectors = []
    all_tables = tableService().list_tables(pos_id)
    if len(all_tables) == 0:
        return sectors

    options = _get_options_sectors_services_period(all_tables, model)
    mask = "NO_SALE_PANEL_IN_SERVICE_PERIOD"
    selected = mw_helper.show_check_list_box_dialog(pos_id, options, "$SELECT_SECTORS_SERVICE_PERIOD", mask=mask)
    if not len(selected):
        return None

    all_store = False
    all_sectors = False
    for selected_index in selected:
        if int(selected_index) == 1:
            all_sectors = True
            continue
        if int(selected_index) == 0:
            all_store = True
            break
        sector = options[int(selected_index)]
        sectors.append(sector.split("##")[0])

    if _display_sector_confirmation(pos_id, all_store, all_sectors, sectors, service_period_name):
        return sectors
    return None


def _get_options_sectors_services_period(all_tables, model):
    # type: (List, eTree.XML) -> List

    options = []
    [options.append(table.sector) for table in all_tables if table.sector not in options]
    options.sort(key=lambda item: item)
    for index, sector in enumerate(options):
        name = DEFAULT_NAME.format(sector)
        data = {"index": index + 1, "id": None, "name": name}
        options[index] = data

    title = translate_message(model, "ALL_STORE")
    name = "{}##True".format(DEFAULT_NAME.format(title))
    insert_option(options, -1, name)

    title = translate_message(model, "ALL_SECTORS")
    name = DEFAULT_NAME.format(title)
    insert_option(options, 0, name)

    options.sort(key=lambda x: (x["index"]))
    return ["{0}".format(c["name"]) for c in options]


def insert_option(options, index, name):
    # type: (List, int, str) -> None

    data = {"index": index, "id": None, "name": name}
    options.insert(index, data)


def _display_sector_confirmation(pos_id, all_store, all_sectors, sectors, service_period_name):
    # type: (int, bool, bool, List[str], str) -> bool

    if all_store:
        message = "$SERVICE_PERIOD_WILL_BE_ACTIVATE_FOR_ENTIRE_STORE|{0}".format(service_period_name)
    elif all_sectors:
        message = "$SERVICE_PERIOD_WILL_BE_ACTIVATE_FOR_ALL_SECTORS|{0}".format(service_period_name)
    else:
        sectors_name = ", ".join(sectors)
        message = "$SERVICE_PERIOD_WILL_BE_ACTIVATE_FOR_SECTORS|{0}|{1}".format(service_period_name, sectors_name)
    return show_confirmation(pos_id, message)



#
# M E S S A G E   T O K E N S
#
# TOKEN FORMAT (32 bits): 0xPGGXNNNN
#    P  - message priority
#    G  - message group
#    X  - not used
#    N  - message number
#

MSGPRT_HIGH = "F"
MSGPRT_MEDHIG = "A"
MSGPRT_MEDLOW = "5"
MSGPRT_LOW = "0"


def create_token(priority, message_grourp, number):
    # type: (str, str, str) -> int
    while len(number) < 4:
        number = "0" + number

    while len(message_grourp) < 2:
        message_grourp = "0" + message_grourp

    return int("0x" + priority + message_grourp + "0" + number, 16)


def get_message_number(token):
    # type: (int) -> int
    return token & 0xFFFF


def get_token_hexa_string(token):
    # type: (int) -> str
    return "0x" + format(token, 'x')


# SAT Module
TK_SAT_PROCESS_REQUEST = create_token(MSGPRT_HIGH, "30", "1")
TK_SAT_STATUS_REQUEST = create_token(MSGPRT_LOW, "30", "2")
TK_SAT_OPERATIONAL_STATUS_REQUEST = create_token(MSGPRT_LOW, "30", "3")
TK_SAT_CANCEL_SALE = create_token(MSGPRT_LOW, "30", "4")

# FiscalWrapper Module
TK_FISCALWRAPPER_PROCESS_REQUEST = create_token(MSGPRT_HIGH, "31", "1")
TK_FISCALWRAPPER_SEARCH_REQUEST = create_token(MSGPRT_LOW, "31", "2")
TK_FISCALWRAPPER_REPRINT = create_token(MSGPRT_HIGH, "31", "3")
TK_FISCALWRAPPER_SEFAZ_CONNECTIVITY = create_token(MSGPRT_HIGH, "31", "4")
TK_FISCALWRAPPER_SAT_OPERATIONAL_STATUS_REQUEST = create_token(MSGPRT_HIGH, "31", "5")
TK_FISCALWRAPPER_EXPORT_FISCAL_FILES = create_token(MSGPRT_LOW, "31", "6")
TK_FISCALWRAPPER_GET_NF_TYPE = create_token(MSGPRT_LOW, "31", "7")
TK_FISCALWRAPPER_SAT_PROCESS_PAYMENT = create_token(MSGPRT_LOW, "31", "8")
TK_FISCALWRAPPER_SAT_PAYMENT_STATUS = create_token(MSGPRT_LOW, "31", "9")
TK_FISCALWRAPPER_CANCEL_ORDER = create_token(MSGPRT_LOW, "31", "10")
TK_SAT_PROCESS_PAYMENT = create_token(MSGPRT_LOW, "31", "11")
TK_SAT_PAYMENT_STATUS = create_token(MSGPRT_LOW, "31", "12")
TK_SAT_CANCEL_ORDER = create_token(MSGPRT_LOW, "31", "13")
TK_FISCALWRAPPER_RE_SIGN_XML = create_token(MSGPRT_LOW, "31", "14")
TK_FISCALWRAPPER_SITUATION = create_token(MSGPRT_LOW, "31", "15")  # 51380245
TK_FISCALWRAPPER_GET_CST = create_token(MSGPRT_LOW, "31", "16")
TK_FISCALWRAPPER_GET_BASE_REDUCTION = create_token(MSGPRT_LOW, "31", "17")
TK_FISCALWRAPPER_GET_FISCAL_XML = create_token(MSGPRT_LOW, "31", "18")
TK_FISCALWRAPPER_GET_IBPT_TAX = create_token(MSGPRT_LOW, "31", "19")
TK_FISCALWRAPPER_GET_ORDERS_DATA = create_token(MSGPRT_LOW, "31", "20")  # 51380246
TK_FISCALWRAPPER_GET_LAST_ORDER_ID = create_token(MSGPRT_LOW, "31", "21")  # 51380247
TK_FISCALWRAPPER_PRINT_DONATION_COUPON = create_token(MSGPRT_LOW, "31", "22")
TK_FISCALWRAPPER_GET_CONTINGENCY_STATUS = create_token(MSGPRT_LOW, "31", "23")
TK_FISCALWRAPPER_CHANGE_CONTINGENCY_STATUS = create_token(MSGPRT_LOW, "31", "24")
TK_FISCALWRAPPER_GET_CERTIFICATE_EXPIRATION_DAYS = create_token(MSGPRT_LOW, "31", "25")
TK_FISCALWRAPPER_GET_TIP = create_token(MSGPRT_LOW, "31", "31")
TK_FISCALWRAPPER_SAT_STORE_STATUS = create_token(
    priority=MSGPRT_LOW,
    message_grourp="31",
    number="32",
)


# SiTEF Module
TK_SITEF_ADD_PAYMENT = create_token(MSGPRT_HIGH, "32", "1")
TK_SITEF_CANCEL_PAYMENT = create_token(MSGPRT_LOW, "32", "2")
TK_SITEF_FINISH_PAYMENT = create_token(MSGPRT_LOW, "32", "3")
TK_SITEF_CONNECTIVITY_TEST = create_token(MSGPRT_LOW, "32", "4")
TK_SITEF_AVAILABLE = create_token(MSGPRT_LOW, "32", "5")
TK_SITEF_CONFIRM_FINISH = create_token(MSGPRT_LOW, "32", "6")
TK_SITEF_NOT_ALL_PAYMENTS_REFUNDED = create_token(MSGPRT_LOW, "32", "1001")
TK_SITEF_REFUND_CANCELED = create_token(MSGPRT_LOW, "32", "1002")
TK_SITEF_GET_DOC = create_token(MSGPRT_LOW, "32", "7")
TK_SITEF_CANCEL_PAYMENT_ORDER = create_token(MSGPRT_LOW, "32", "8")

# FingerPrint Reader
TK_FPR_ENROLL_USER = create_token(MSGPRT_LOW, "33", "1")
TK_FPR_IDENTIFY_USER = create_token(MSGPRT_LOW, "33", "2")
TK_FPR_AUTHORIZE_USER = create_token(MSGPRT_LOW, "33", "3")
TK_FPR_OK = create_token(MSGPRT_LOW, "33", "4")
TK_FPR_RE_ENROLL_USER = create_token(MSGPRT_LOW, "33", "5")
TK_FPR_TIMEOUT = create_token(MSGPRT_LOW, "33", "6")
TK_FPR_FINGER_PRINT_ALREADY_REGISTERED = create_token(MSGPRT_LOW, "33", "7")

# KDSMonitor Module
TK_KDSMONITOR_STATUS_REQUEST = create_token(MSGPRT_LOW, "34", "1")

# Maintenance Module
TK_MAINTENANCE_TERMINATE_REQUEST = create_token(MSGPRT_LOW, "35", "1")
TK_MAINTENANCE_RESTART_REQUEST = create_token(MSGPRT_LOW, "35", "2")
TK_MAINTENANCE_RECREATE_FILES_XML = create_token(MSGPRT_LOW, "35", "3")  # token 55574531
TK_MAINTENANCE_RECREATE_FILE_NUMBER = create_token(MSGPRT_LOW, "35", "4")
TK_MAINTENANCE_CLEAN_AUDIT_LOG = create_token(MSGPRT_LOW, "35", "5")
TK_MAINTENANCE_DATABASE_CLEANUP_STATUS = create_token(MSGPRT_LOW, "35", "6")

# Chat Controller
TK_CHAT_SEND_MESSAGE = create_token(MSGPRT_LOW, "36", "1")
TK_CHAT_GET_UPDATES = create_token(MSGPRT_LOW, "36", "2")
TK_CHAT_MARK_UPDATES_RECEIVED = create_token(MSGPRT_LOW, "36", "3")
TK_CHAT_GET_LAST_MESSAGES = create_token(MSGPRT_LOW, "36", "4")
TK_CHAT_INVALID_REQUEST = create_token(MSGPRT_LOW, "36", "100")

# Remote Order
TK_REMOTE_ORDER_GET_STORED_ORDERS = create_token(MSGPRT_LOW, "37", "1")
TK_REMOTE_ORDER_SEND_ORDER_TO_PRODUCTION = create_token(MSGPRT_LOW, "37", "2")
TK_REMOTE_ORDER_GET_STORE = create_token(MSGPRT_LOW, "37", "3")
TK_REMOTE_ORDER_CLOSE_STORE = create_token(MSGPRT_LOW, "37", "4")
TK_REMOTE_ORDER_OPEN_STORE = create_token(MSGPRT_LOW, "37", "5")
TK_REMOTE_ORDER_INVALID_ORDER_ID = create_token(MSGPRT_LOW, "37", "100")
TK_REMOTE_STORE_ALREADY_OPEN = create_token(MSGPRT_LOW, "37", "101")
TK_REMOTE_STORE_ALREADY_CLOSED = create_token(MSGPRT_LOW, "37", "102")
TK_REMOTE_ORDER_REPRINT = create_token(MSGPRT_LOW, "37", "104")
TK_REMOTE_ORDER_GET_AVAILABLE_PRODUCTS = create_token(MSGPRT_LOW, "37", "201")
TK_REMOTE_ORDER_GET_UNAVAILABLE_PRODUCTS = create_token(MSGPRT_LOW, "37", "202")
TK_REMOTE_ORDER_UPDATE_AVAILABLE_PRODUCTS = create_token(MSGPRT_LOW, "37", "203")
TK_REMOTE_ORDER_ERROR = create_token(MSGPRT_LOW, "37", "204")
TK_REMOTE_ORDER_CHECK_RUPTURA_DIFF_ITEMS = create_token(MSGPRT_LOW, "37", "205")
TK_REMOTE_ORDER_VOID_REMOTE_ORDER = create_token(MSGPRT_LOW, "37", "206")
TK_REMOTE_ORDER_CREATE_AND_PRODUCE = create_token(MSGPRT_LOW, "37", "207")
TK_REMOTE_ORDER_CREATE = create_token(MSGPRT_LOW, "37", "208")
TK_REMOTE_ORDER_GET_VOIDED_ORDERS = create_token(MSGPRT_LOW, "37", "209")
TK_REMOTE_ORDER_GET_POS_ID = create_token(MSGPRT_LOW, "37", "210")
TK_REMOTE_ORDER_GET_CONFIRMED_ORDERS = create_token(MSGPRT_LOW, "37", "211")
TK_REMOTE_ORDER_GET_RUPTURED_ITEMS = create_token(MSGPRT_LOW, "37", "213")
TK_REMOTE_ORDER_GET_STORE_STATUS = create_token(MSGPRT_LOW, "37", "214")
TK_REMOTE_ORDER_CHECK_IF_ORDER_EXISTS = create_token(MSGPRT_LOW, "37", "215")
TK_REMOTE_ORDER_REPRINT_DELIVERY = create_token(MSGPRT_LOW, "37", "216")
TK_REMOTE_ORDER_GET_LOGISTIC_PARTNERS = create_token(MSGPRT_LOW, "37", "217")
TK_REMOTE_ORDER_SEARCH_LOGISTIC = create_token(MSGPRT_LOW, "37", "218")
TK_REMOTE_ORDER_SEND_ORDER_TO_LOGISTIC = create_token(MSGPRT_LOW, "37", "219")
TK_REMOTE_ORDER_CANCEL_LOGISTIC = create_token(MSGPRT_LOW, "37", "220")
TK_REMOTE_ORDER_SAVE_ORDER = create_token(MSGPRT_LOW, "37", "221")
TK_REMOTE_ORDER_GET_DELIVERY_FEE_CODE = create_token(MSGPRT_LOW, "37", "222")
TK_REMOTE_ORDER_CONFIRM_DELIVERY_PAYMENT = create_token(MSGPRT_LOW, "37", "223")
TK_REMOTE_ORDER_INFORM_ORDER_READY = create_token(MSGPRT_LOW, "37", "225")
TK_REMOTE_ORDER_GET_DELIVERY_BRANDS = create_token(MSGPRT_LOW, "37", "226")


# Daily Goals
TK_DAILY_GOALS_UPDATE_GOALS = create_token(MSGPRT_LOW, "38", "1")
TK_DAILY_GOALS_UPDATE_TOTAL_SOLD = create_token(MSGPRT_LOW, "38", "2")
TK_DAILY_GOALS_TERMINATE_REQUEST = create_token(MSGPRT_LOW, "38", "3")
TK_DAILY_GOALS_GET_SALES_CHALLENGE = create_token(MSGPRT_LOW, "38", "4")
TK_DAILY_GOALS_INSERT_SALES_CHALLENGE = create_token(MSGPRT_LOW, "38", "5")
TK_DAILY_GOALS_DELETE_SALES_CHALLENGE = create_token(MSGPRT_LOW, "38", "6")
TK_DAILY_GOALS_UPDATE_SALES_CHALLENGE = create_token(MSGPRT_LOW, "38", "7")

# Pickup Display Module
TK_PICKUPDISPLAY_ADD_READY_TO_PICK_ORDER = create_token(MSGPRT_LOW, "39", "1")
TK_PICKUPDISPLAY_ADD_PAID_ORDER = create_token(MSGPRT_LOW, "39", "2")
TK_PICKUPDISPLAY_ADD_PICKED_UP_ORDER = create_token(MSGPRT_LOW, "39", "3")
TK_PICKUPDISPLAY_UPDATE_ORDER = create_token(MSGPRT_LOW, "39", "4")
TK_PICKUPDISPLAY_SEND_CACHED_ORDERS = create_token(MSGPRT_LOW, "39", "5")

# Custom Production
TK_PROD_ENABLEVIEW = create_token(MSGPRT_LOW, "0C", "100")
TK_PROD_DISABLEVIEW = create_token(MSGPRT_LOW, "0C", "101")
TK_PROD_PURGE = create_token(MSGPRT_LOW, "0C", "103")
TK_PROD_GET_ORDER_MAX_PREP_TIME = create_token(MSGPRT_LOW, "C", "3")
TK_PROD_GET_PRINTERS_INFO = create_token(MSGPRT_LOW, "C", "201")
TK_PROD_GET_COURSES = create_token(MSGPRT_LOW, "C", "202")

TK_PROD_ORDER_PICKLIST = create_token(MSGPRT_LOW, "C", "301")

# BK Office uploader
TK_BKOFFICEUPLOADER_UPDATE_ORDERPICTURE = create_token(MSGPRT_LOW, "40", "1")
TK_BKOFFICEUPLOADER_SEND_SANGRIA = create_token(MSGPRT_LOW, "40", "2")
TK_BKOFFICEUPLOADER_SEND_BKOFFICE_ORDER = create_token(MSGPRT_LOW, "40", "3")

# Sapiens uploader
TK_SAPIENS_GENERATE_SALES = create_token(MSGPRT_LOW, "41", "1")
TK_SAPIENS_GENERATE_REDUCTIONZ = create_token(MSGPRT_LOW, "41", "2")
TK_SAPIENS_GENERATE_SALE_CENTER = create_token(MSGPRT_LOW, "41", "3")
TK_SAPIENS_UPLOAD = create_token(MSGPRT_LOW, "41", "4")
TK_SAPIENS_GENERATE_ALL_FILES_PER_PERIOD = create_token(MSGPRT_LOW, "41", "5")
TK_SAPIENS_PERFORM_BACKUP = create_token(MSGPRT_LOW, "41", "6")
TK_SAPIENS_PERFORM_FISCAL_VALIDATION = create_token(MSGPRT_LOW, "41", "7")

# Ruptura messages
TK_RUPTURA_GET_ENABLED = create_token(MSGPRT_LOW, "42", "3")
TK_RUPTURA_GET_DISABLED = create_token(MSGPRT_LOW, "42", "4")
TK_RUPTURA_GET_ENABLED_PRODUCTS = create_token(MSGPRT_LOW, "42", "5")
TK_RUPTURA_GET_DISABLED_PRODUCTS = create_token(MSGPRT_LOW, "42", "6")
TK_RUPTURA_IS_PROCESSED = create_token(MSGPRT_LOW, "42", "7")
TK_RUPTURA_UPDATE_ITEMS = create_token(MSGPRT_LOW, "42", "8")
TK_RUPTURA_CHECK_ITEMS = create_token(MSGPRT_LOW, "42", "9")
TK_RUPTURA_GET_CLEAN_TIME_WINDOW = create_token(MSGPRT_LOW, "42", "10")
TK_RUPTURA_CLEAN_WITH_CATALOG_UPDATE = create_token(MSGPRT_LOW, "42", "11")
TK_RUPTURA_GET_DAILY_DISHES = create_token(MSGPRT_LOW, "42", "12")
TK_RUPTURA_UPDATE_DAILY_DISHES = create_token(MSGPRT_LOW, "42", "13")

# Blacklist
TK_BLACKLIST_CHECK_STRING = create_token(MSGPRT_LOW, "44", "0")
TK_BLACKLIST_FILTER_STRING = create_token(MSGPRT_LOW, "44", "1")

# PafEcfListener
TK_VERIFY_PAF_ECF_LISTENER = create_token(MSGPRT_LOW, "45", "1")
TK_POS_FUNCTION_CHANGED = create_token(MSGPRT_LOW, "45", "2")

# Pos Updater
TK_POS_UPDATER_PERFORM_CATALOG_UPDATE = create_token(MSGPRT_LOW, "95", "1")
TK_POS_UPDATER_PERFORM_MEDIA_UPDATE = create_token(MSGPRT_LOW, "95", "2")
TK_POS_UPDATER_PERFORM_USER_UPDATE = create_token(MSGPRT_LOW, "95", "3")
TK_POS_GET_CATALOG_VERSION_TO_APPLY = create_token(MSGPRT_LOW, "95", "4")
TK_POS_UPDATER_PERFORM_LOADER_UPDATE = create_token(MSGPRT_LOW, "95", "5")
TK_POS_UPDATER_GET_VERSIONS = create_token(MSGPRT_LOW, "95", "6")
TK_POS_UPDATER_PERFORM_BINARIES_UPDATE = create_token(MSGPRT_LOW, "95", "7")
TK_POS_HAS_BINARY_UPDATE_TO_APPLY = create_token(MSGPRT_LOW, "95", "8")
TK_POS_UPDATER_PERFORM_CHROME_UPDATE = create_token(MSGPRT_LOW, "95", "9")

# Discount
TK_DISCOUNT_APPLY = create_token(MSGPRT_HIGH, "43", "1")
TK_DISCOUNT_CLEAR = create_token(MSGPRT_HIGH, "43", "2")

# NDiscount
TK_NDISCOUNT_ADD_BENEFIT = create_token(MSGPRT_LOW, "112", "1")
TK_NDISCOUNT_REMOVE = create_token(MSGPRT_LOW, "112", "2")
TK_NDISCOUNT_GET_BENEFITS = create_token(MSGPRT_LOW, "112", "7")
TK_NDISCOUNT_ON_ITEM_SALE = create_token(MSGPRT_LOW, "112", "8")
TK_NDISCOUNT_ON_ORDER_TOTAL = create_token(MSGPRT_LOW, "112", "9")
TK_NDISCOUNT_ON_ORDER_REOPEN = create_token(MSGPRT_LOW, "112", "10")
TK_NDISCOUNT_ON_TABLE_TOTAL = create_token(MSGPRT_LOW, "112", "11")
TK_NDISCOUNT_ON_TABLE_REOPEN = create_token(MSGPRT_LOW, "112", "12")
TK_NDISCOUNT_ON_TABLE_OPEN = create_token(MSGPRT_LOW, "112", "13")
TK_NDISCOUNT_ON_TAB_OPEN = create_token(MSGPRT_LOW, "112", "14")

# Coupon
TK_NCOUPON_APPLY_COUPON = create_token(MSGPRT_LOW, "113", "1")
TK_NCOUPON_UNLOCK_COUPON = create_token(MSGPRT_LOW, "113", "2")
TK_NCOUPON_BURN_COUPON = create_token(MSGPRT_LOW, "113", "3")
TK_NCOUPON_GET_COUPON_INFO = create_token(MSGPRT_LOW, "113", "4")
TK_NCOUPON_TEST_CONNECTION = create_token(MSGPRT_LOW, "113", "5")

# Electronic Payment
TK_ELECTRONIC_PAYMENT_ADD_PAYMENT = create_token(MSGPRT_LOW, "E", "1")
TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT = create_token(MSGPRT_LOW, "E", "2")
TK_ELECTRONIC_PAYMENT_FINISH_PAYMENT = create_token(MSGPRT_LOW, "E", "3")
TK_ELECTRONIC_PAYMENT_CONNECTIVITY_TEST = create_token(MSGPRT_LOW, "E", "4")
TK_ELECTRONIC_PAYMENT_SEND_RESPONSE = create_token(MSGPRT_LOW, "E", "5")
TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT_ORDER = create_token(MSGPRT_LOW, "E", "6")
TK_ELECTRONIC_PAYMENT_CONFIRM_FINISH = create_token(MSGPRT_LOW, "E", "7")
TK_ELECTRONIC_PAYMENT_GET_ELECTRONIC_TYPES = create_token(MSGPRT_LOW, "E", "8")
TK_ELECTRONIC_PAYMENT_GET_DOCUMENT = create_token(MSGPRT_LOW, "E", "9")
TK_ELECTRONIC_PAYMENT_NOT_ALL_PAYMENTS_REFUNDED = create_token(MSGPRT_LOW, "E", "1001")
TK_ELECTRONIC_PAYMENT_REFUND_CANCELED = create_token(MSGPRT_LOW, "E", "1002")

# Nsale
TK_SLCTRL_OMGR_UPDATEORDERTENDERDETAIL=create_token(MSGPRT_LOW, "A", "32")

# Device Manager
TK_GET_USED_SERVICE = create_token(MSGPRT_LOW, "D", "1")

# Loyalty
TK_LOYALTY_GET_LOYALTY_ID = create_token(MSGPRT_LOW, "111", "1")
TK_LOYALTY_GET_AND_LOCK_VOUCHER = create_token(MSGPRT_LOW, "111", "2")
TK_LOYALTY_BURN_VOUCHER = create_token(MSGPRT_LOW, "111", "3")
TK_LOYALTY_UNLOCK_VOUCHER = create_token(MSGPRT_LOW, "111", "4")
TK_LOYALTY_IS_AVAILABLE = create_token(MSGPRT_LOW, "111", "FF")



import logging

import sysactions

from xml.etree import cElementTree as eTree
from typing import (
    Optional,
    Dict,
    List,
)

from mw_helper import show_filtered_list_box_dialog, show_numpad_dialog, show_message_dialog
from sysactions import show_messagebox, get_current_order, StopAction, get_model, get_posot, show_keyboard

from actions.models import PaymentType
from actions.coupon.apply_coupon import apply_coupon
from actions.discount.get_discount import get_visible_discounts
from actions.util import (
    get_authorization,
    get_user_level,
)
from application.interactor import NDiscountInteractor
from manager.reports import justification_report
from model.customexception import ErrorAddingBenefits, DiscountMinLevelException, ErrorApplyingDiscount

logger = logging.getLogger("PosActions")


def apply_discount(
        pos_id,                     # type: str
        selected_coupon=None,       # type: Optional[Dict]
        amount=0.0,                 # type: Optional[float]
        user_auth_level=-1,         # type: Optional[int]
        user_auth_name=None,        # type: Optional[str]
        user_auth_id=None,          # type: Optional[str]
        justification=None,         # type: Optional[str]
):
    # type: (...) -> bool

    try:
        coupons = get_visible_discounts(pos_id)
        selected_coupon = _select_coupon(pos_id, selected_coupon, coupons)
        if not selected_coupon:
            return False

        coupon = selected_coupon.get("coupon", {})
        model = get_model(pos_id)

        if user_auth_id is None:
            operator = model.find("Operator")
            user_auth_id = operator.get("id")

        try:
            if coupon.get("requestJustification", False):
                if justification is None:
                    justification = _get_justification_discount(pos_id=pos_id)

                coupon['justification'] = justification
                coupon['appliedBy'] = user_auth_id

            if coupon.get("askValue", False) and not amount:
                coupon["index"] = _get_discount_type(coupon)
                amount = amount or _get_discount_amount(coupon, pos_id)
                coupon = set_coupon_amount(
                    coupon=coupon,
                    amount=amount,
                )

            _add_benefit(
                pos_id=pos_id,
                coupon=coupon,
                user_auth_level=user_auth_level,
                benefit_id=selected_coupon.get("benefit_id", ""),
                user_auth=user_auth_name,
                justification=justification,
                amount=amount,
            )
            return True
        except DiscountMinLevelException as _:
            show_messagebox(pos_id, "$NDISCOUNT_ERROR_USER_MIN_LEVEL")
            user_was_authorized, user_auth = get_authorization(
                pos_id=pos_id,
                min_level=coupon.get("conditions").get("userMinLevel", 0),
                model=model
            )
            if user_was_authorized:
                user_xml_str = sysactions.get_user_information(user_auth)
                user_element = eTree.XML(user_xml_str).find("user")
                user_auth_level = int(user_element.attrib["Level"])
                user_auth_id = str(user_element.attrib["UserId"])
                coupon['authorizedBy'] = user_auth_id

                apply_discount(
                    pos_id=pos_id,
                    selected_coupon=selected_coupon,
                    amount=amount,
                    user_auth_level=user_auth_level,
                    user_auth_name=user_element.attrib["UserName"],
                    user_auth_id=user_auth_id,
                    justification=justification
                )
                return True

    except (ErrorAddingBenefits, ErrorApplyingDiscount) as ex:
        logger.warn("Error applying discount ${}".format(ex.message))
        show_message_dialog(pos_id, "$ERROR", "${}".format(ex.message))
    except Exception as ex:
        logger.exception("Error applying discount")
        message = "FATAL_ERROR"
        if hasattr(ex, 'message'):
            message = "${}".format(message)
        show_message_dialog(pos_id, "$ERROR", message)

    return False


def _select_coupon(
    pos_id,             # type: str
    selected_coupon,    # type: Optional[Dict]
    coupons,            # type: List
):
    # type: (...) -> Optional[Dict]

    if not selected_coupon:
        if len(coupons) == 0:
            show_messagebox(pos_id, "$DISCOUNT_NOT_FOUND", icon="warning")
            return None

        selected_index = _select_discount(pos_id, coupons)
        if selected_index is None:
            return None

        selected_coupon = filter(lambda c: c["index"] == selected_index + 1, coupons)
        if not selected_coupon:
            return None

        selected_coupon = selected_coupon[0]

    return selected_coupon


def _get_discount_amount(coupon, pos_id):
    model = get_model(pos_id)
    order = get_current_order(model)
    amount, percentage = _ask_discount_amount(pos_id, order, coupon=coupon)
    if percentage is not None:
        amount = percentage

    return amount


def apply_amount_discount(pos_id, pos_ot, order, coupon=None, discount_type=None, discount_value=None):
    amount, _ = _ask_discount_amount(pos_id, order, coupon, discount_type, discount_value)
    pos_ot.applyDiscount(1, amount, forcebyitem=1)


def _ask_discount_amount(pos_id, order, coupon=None, discount_type=None, discount_value=None):
    if not discount_type:
        discount_type = coupon['index']
        title = "$DISCOUNT_VALUE_TITLE" if discount_type == PaymentType.AMOUNT.value else "$DISCOUNT_PERCENTAGE_TITLE"
        mask = "CURRENCY" if discount_type == PaymentType.AMOUNT.value else "NUMBER"
        if not discount_value:
            discount_value = show_numpad_dialog(pos_id, title, "", mask)
            if discount_value in [None, '']:
                raise StopAction()

    if discount_type == PaymentType.AMOUNT.value:
        return discount_value, None

    amount = float(order.get("totalGross"))

    return (amount * discount_value) / 100, discount_value


def _select_discount(pos_id, coupons):
    options = ["{0} - {1}".format(coupon["index"], coupon["name"].encode("UTF-8")) for coupon in coupons]
    return show_filtered_list_box_dialog(pos_id, options, "$SELECT_ONE_COUPON", mask="NOFILTER")


def _add_benefit(
    pos_id,             # type: str
    coupon,             # type: Dict
    user_auth_level,    # type: int
    benefit_id,         # type: str
    user_auth,          # type: str
    justification,      # type: str
    amount,             # type: float
):
    # type: (...) -> None

    authorization_level = int(coupon.get("conditions", {}).get("authorizationLevel", 0))
    if _get_discount_authorization(
        pos_id=pos_id,
        coupon=coupon,
        authorization_level=authorization_level,
    ) is False:
        raise StopAction()

    if justification or amount:
        NDiscountInteractor.add_benefit(
            pos_id=pos_id,
            benefit_id="",
            benefit=coupon,
            user_auth_level=user_auth_level,
        )
        if justification and coupon.get("printJustifyCoupon", False):
            justification_report(
                pos_id=pos_id,
                justification=justification,
                discount_description=coupon.get('userDescription', ""),
                authorized_by=user_auth,
            )
        return

    return apply_coupon(
        pos_id=pos_id,
        table_id=None,
        coupon_code=benefit_id,
        user_auth_level=user_auth_level,
        raise_error=True,
    )


def _get_object_json(json_object, key, value):
    return [obj for obj in json_object if obj[key] == value][0]


def set_coupon_amount(coupon, amount):
    if coupon.get("actions", False) and coupon.get("actions").get("0", False):
        action = coupon.get("actions").get("0")
        value_actions = ["discountAmountOnTotal", "discountPercentageOnTotal"]
        group_actions = ["discountPercentageByGroupsOnTotal", "discountAmountByGroupsOnTotal"]
        if any(x for x in value_actions if x in action.keys()):
            action_name = action.keys()[0]
            action[action_name] = amount
            return coupon
        elif any(x for x in group_actions if x in action.keys()):
            action.values()[0][0]["amount"] = amount
            return coupon

    raise StopAction()


def _get_discount_type(coupon):
    if coupon.get("actions", False) and coupon.get("actions").get("0", False):
        action = coupon.get("actions").get("0")
        amount_actions = ["discountAmountOnTotal", "discountAmountByGroupsOnTotal"]
        percentage_actions = ["discountPercentageOnTotal", "discountPercentageByGroupsOnTotal"]
        if any(x for x in amount_actions if x in action.keys()):
            return PaymentType.AMOUNT.value
        elif any(x for x in percentage_actions if x in action.keys()):
            return PaymentType.PERCENTAGE.value

    raise StopAction()


def _get_justification_discount(pos_id):
    # type: (str) -> str

    while True:
        justification = show_keyboard(pos_id, "$JUSTIFICATION_DISCOUNT", '', timeout=1800000, buttons="$OK")
        if justification is None:
            raise StopAction()

        justification = justification.strip()
        if len(justification) < 5:
            show_messagebox(
                posid=pos_id,
                message="$MINIMUM_5_CHARACTERS_REQUIRED",
            )
            continue

        break

    return justification


def _get_discount_authorization(
    pos_id,                 # type: str
    coupon,                 # type: dict
    authorization_level,    # type: int
):
    # type: (...) -> Optional[dict, bool]

    model = get_model(posid=pos_id)
    operator = model.find("./Operator")
    user_id = operator.get("id")
    current_user_level = get_user_level(user_id=user_id)

    if current_user_level >= authorization_level:
        coupon['authorized_by'] = user_id
        return coupon

    user_was_authorized, user_auth = get_authorization(
        pos_id=pos_id,
        min_level=authorization_level,
        model=model,
    )
    authorized_user_level = get_user_level(user_id=user_auth)
    if user_was_authorized and authorized_user_level >= authorization_level:
        coupon['authorized_by'] = user_auth
        return coupon

    show_messagebox(
        posid=pos_id,
        message="$NDISCOUNT_ERROR_USER_MIN_LEVEL",
    )
    return False



from typing import (
    Any,
    List,
    Optional,
)


class ItemGoals(object):
    def __init__(
        self,
        name,                           # type: unicode
        part_codes,                     # type: List
        part_codes_set,                 # type: List
        quantity_goal,                  # type: int
        quantity_operator_goal,         # type: int
        quantity_sold,                  # type: int
        total_goal_percentage=None,     # type: Optional[float]
        part_code_set_percentage=None,  # type: Optional[float]
    ):
        # type: (...) -> None

        self.name = name
        self.part_codes = part_codes
        self.part_codes_set = part_codes_set
        self.quantity_goal = quantity_goal
        self.quantity_operator_goal = quantity_operator_goal
        self.quantity_sold = quantity_sold
        self.total_goal_percentage = total_goal_percentage
        self.part_code_set_percentage = part_code_set_percentage

    def __getitem__(
        self,
        key,    # type: str
    ):
        # type: (...) -> Any

        return self.__getattribute__(key)



data = (
'ddwim',    # 0x00
'ddwib',    # 0x01
'ddwibs',    # 0x02
'ddwis',    # 0x03
'ddwiss',    # 0x04
'ddwing',    # 0x05
'ddwij',    # 0x06
'ddwic',    # 0x07
'ddwik',    # 0x08
'ddwit',    # 0x09
'ddwip',    # 0x0a
'ddwih',    # 0x0b
'ddyu',    # 0x0c
'ddyug',    # 0x0d
'ddyugg',    # 0x0e
'ddyugs',    # 0x0f
'ddyun',    # 0x10
'ddyunj',    # 0x11
'ddyunh',    # 0x12
'ddyud',    # 0x13
'ddyul',    # 0x14
'ddyulg',    # 0x15
'ddyulm',    # 0x16
'ddyulb',    # 0x17
'ddyuls',    # 0x18
'ddyult',    # 0x19
'ddyulp',    # 0x1a
'ddyulh',    # 0x1b
'ddyum',    # 0x1c
'ddyub',    # 0x1d
'ddyubs',    # 0x1e
'ddyus',    # 0x1f
'ddyuss',    # 0x20
'ddyung',    # 0x21
'ddyuj',    # 0x22
'ddyuc',    # 0x23
'ddyuk',    # 0x24
'ddyut',    # 0x25
'ddyup',    # 0x26
'ddyuh',    # 0x27
'ddeu',    # 0x28
'ddeug',    # 0x29
'ddeugg',    # 0x2a
'ddeugs',    # 0x2b
'ddeun',    # 0x2c
'ddeunj',    # 0x2d
'ddeunh',    # 0x2e
'ddeud',    # 0x2f
'ddeul',    # 0x30
'ddeulg',    # 0x31
'ddeulm',    # 0x32
'ddeulb',    # 0x33
'ddeuls',    # 0x34
'ddeult',    # 0x35
'ddeulp',    # 0x36
'ddeulh',    # 0x37
'ddeum',    # 0x38
'ddeub',    # 0x39
'ddeubs',    # 0x3a
'ddeus',    # 0x3b
'ddeuss',    # 0x3c
'ddeung',    # 0x3d
'ddeuj',    # 0x3e
'ddeuc',    # 0x3f
'ddeuk',    # 0x40
'ddeut',    # 0x41
'ddeup',    # 0x42
'ddeuh',    # 0x43
'ddyi',    # 0x44
'ddyig',    # 0x45
'ddyigg',    # 0x46
'ddyigs',    # 0x47
'ddyin',    # 0x48
'ddyinj',    # 0x49
'ddyinh',    # 0x4a
'ddyid',    # 0x4b
'ddyil',    # 0x4c
'ddyilg',    # 0x4d
'ddyilm',    # 0x4e
'ddyilb',    # 0x4f
'ddyils',    # 0x50
'ddyilt',    # 0x51
'ddyilp',    # 0x52
'ddyilh',    # 0x53
'ddyim',    # 0x54
'ddyib',    # 0x55
'ddyibs',    # 0x56
'ddyis',    # 0x57
'ddyiss',    # 0x58
'ddying',    # 0x59
'ddyij',    # 0x5a
'ddyic',    # 0x5b
'ddyik',    # 0x5c
'ddyit',    # 0x5d
'ddyip',    # 0x5e
'ddyih',    # 0x5f
'ddi',    # 0x60
'ddig',    # 0x61
'ddigg',    # 0x62
'ddigs',    # 0x63
'ddin',    # 0x64
'ddinj',    # 0x65
'ddinh',    # 0x66
'ddid',    # 0x67
'ddil',    # 0x68
'ddilg',    # 0x69
'ddilm',    # 0x6a
'ddilb',    # 0x6b
'ddils',    # 0x6c
'ddilt',    # 0x6d
'ddilp',    # 0x6e
'ddilh',    # 0x6f
'ddim',    # 0x70
'ddib',    # 0x71
'ddibs',    # 0x72
'ddis',    # 0x73
'ddiss',    # 0x74
'dding',    # 0x75
'ddij',    # 0x76
'ddic',    # 0x77
'ddik',    # 0x78
'ddit',    # 0x79
'ddip',    # 0x7a
'ddih',    # 0x7b
'ra',    # 0x7c
'rag',    # 0x7d
'ragg',    # 0x7e
'rags',    # 0x7f
'ran',    # 0x80
'ranj',    # 0x81
'ranh',    # 0x82
'rad',    # 0x83
'ral',    # 0x84
'ralg',    # 0x85
'ralm',    # 0x86
'ralb',    # 0x87
'rals',    # 0x88
'ralt',    # 0x89
'ralp',    # 0x8a
'ralh',    # 0x8b
'ram',    # 0x8c
'rab',    # 0x8d
'rabs',    # 0x8e
'ras',    # 0x8f
'rass',    # 0x90
'rang',    # 0x91
'raj',    # 0x92
'rac',    # 0x93
'rak',    # 0x94
'rat',    # 0x95
'rap',    # 0x96
'rah',    # 0x97
'rae',    # 0x98
'raeg',    # 0x99
'raegg',    # 0x9a
'raegs',    # 0x9b
'raen',    # 0x9c
'raenj',    # 0x9d
'raenh',    # 0x9e
'raed',    # 0x9f
'rael',    # 0xa0
'raelg',    # 0xa1
'raelm',    # 0xa2
'raelb',    # 0xa3
'raels',    # 0xa4
'raelt',    # 0xa5
'raelp',    # 0xa6
'raelh',    # 0xa7
'raem',    # 0xa8
'raeb',    # 0xa9
'raebs',    # 0xaa
'raes',    # 0xab
'raess',    # 0xac
'raeng',    # 0xad
'raej',    # 0xae
'raec',    # 0xaf
'raek',    # 0xb0
'raet',    # 0xb1
'raep',    # 0xb2
'raeh',    # 0xb3
'rya',    # 0xb4
'ryag',    # 0xb5
'ryagg',    # 0xb6
'ryags',    # 0xb7
'ryan',    # 0xb8
'ryanj',    # 0xb9
'ryanh',    # 0xba
'ryad',    # 0xbb
'ryal',    # 0xbc
'ryalg',    # 0xbd
'ryalm',    # 0xbe
'ryalb',    # 0xbf
'ryals',    # 0xc0
'ryalt',    # 0xc1
'ryalp',    # 0xc2
'ryalh',    # 0xc3
'ryam',    # 0xc4
'ryab',    # 0xc5
'ryabs',    # 0xc6
'ryas',    # 0xc7
'ryass',    # 0xc8
'ryang',    # 0xc9
'ryaj',    # 0xca
'ryac',    # 0xcb
'ryak',    # 0xcc
'ryat',    # 0xcd
'ryap',    # 0xce
'ryah',    # 0xcf
'ryae',    # 0xd0
'ryaeg',    # 0xd1
'ryaegg',    # 0xd2
'ryaegs',    # 0xd3
'ryaen',    # 0xd4
'ryaenj',    # 0xd5
'ryaenh',    # 0xd6
'ryaed',    # 0xd7
'ryael',    # 0xd8
'ryaelg',    # 0xd9
'ryaelm',    # 0xda
'ryaelb',    # 0xdb
'ryaels',    # 0xdc
'ryaelt',    # 0xdd
'ryaelp',    # 0xde
'ryaelh',    # 0xdf
'ryaem',    # 0xe0
'ryaeb',    # 0xe1
'ryaebs',    # 0xe2
'ryaes',    # 0xe3
'ryaess',    # 0xe4
'ryaeng',    # 0xe5
'ryaej',    # 0xe6
'ryaec',    # 0xe7
'ryaek',    # 0xe8
'ryaet',    # 0xe9
'ryaep',    # 0xea
'ryaeh',    # 0xeb
'reo',    # 0xec
'reog',    # 0xed
'reogg',    # 0xee
'reogs',    # 0xef
'reon',    # 0xf0
'reonj',    # 0xf1
'reonh',    # 0xf2
'reod',    # 0xf3
'reol',    # 0xf4
'reolg',    # 0xf5
'reolm',    # 0xf6
'reolb',    # 0xf7
'reols',    # 0xf8
'reolt',    # 0xf9
'reolp',    # 0xfa
'reolh',    # 0xfb
'reom',    # 0xfc
'reob',    # 0xfd
'reobs',    # 0xfe
'reos',    # 0xff
)



import json

from typing import Dict

from ._Message import Message
from ._DataType import DataType
from ._DefaultToken import DefaultToken


class SwaggerMessage(Message):
    def __init__(self, status_code, body=None, headers=None):
        # type: (int, str, Dict[str, str]) -> None
        self.status_code = status_code
        self.body = body
        self.headers = headers

        super(SwaggerMessage, self).__init__(DefaultToken.TK_SYS_ACK, DataType.swagger, b"")



# coding=utf-8
import datetime
import json
import os
import re
import zlib

from logging import Logger
from string import upper
from xml.etree import cElementTree as eTree

import msgbus
import sysactions

from interactor.route.order.add_items.dto import ToAddItem
from interactor.route.order.change_items_quantity.dto import ToChangeItems
from model.error_type import ExternalServicesErrorsType, PosValidationErrorsType
from model.exception import ExternalServiceErrors, ValidationError
from model.config import Config
from model.repository import PosRepositoryModel
from model.states import OperatorStates, PosStates, OrderStates
from model.tokens import BusTokens
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message
from msgbus import TK_SYS_ACK
from posot import OrderTaker
from sysactions import read_scanner, get_used_service, show_messagebox, get_current_order, print_text
from typing import Optional, Any, Dict, List
from mwhelper import ModelService
from services import NavigationService

currentOrder = "CurrentOrder/Order"


class PosRepository(PosRepositoryModel):

    def __init__(
        self,
        message_bus,         # type: MbContextMessageBus
        logger,              # type: Logger
        config,              # type: Config
        navigation_service,  # type: NavigationService
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.logger = logger
        self.config = config
        self.navigation_service = navigation_service
        self.get_nf_type()

        self.languages_translations = {
            "EN": "en-us",
            "ES": "es-es",
            "PT": "pt-br",
        }

        self.descriptions = None

    def get_model(self, pos_id):
        # type: (int) -> eTree

        return sysactions.get_model(pos_id)

    def get_pos_ot(self, model):
        # type: (eTree) -> OrderTaker

        return sysactions.get_posot(model)

    def get_price_list(self, model):
        # type: (eTree) -> str

        pos_id = model.get("posId")
        model_service = ModelService(pos_id=pos_id)
        return model_service.get_price_list_from_model()

    def is_day_opened(self, model):
        # type: (eTree) -> bool

        return model.find("PosState").get("state") == PosStates.OPENED

    def is_day_blocked(self, model):
        # type: (eTree) -> bool

        return model.find("PosState").get("state") == PosStates.BLOCKED

    def is_operator_logged(self, model):
        # type: (eTree) -> bool

        for op in model.findall("Operator"):
            if op.get("state") == OperatorStates.LOGGEDIN:
                return True
        return False

    def translate_message(
        self,
        model,      # type: eTree
        message,    # type: str
    ):
        # type: (...) -> str

        return sysactions.translate_message(
            model=model,
            message_key=message,
        )

    def has_operator_opened(self, model):
        # type: (eTree) -> bool

        for op in model.findall("Operators/Operator"):
            if op.get("state") in [OperatorStates.LOGGEDIN, OperatorStates.OPENED]:
                return True
        return False

    def get_current_operator(self, model):
        # type: (eTree) -> Optional[eTree]

        for op in model.findall("Operator"):
            if op.get("state") == OperatorStates.LOGGEDIN:
                return op
        return None

    def has_current_order(self, model):
        # type: (eTree) -> bool

        order = model.find(currentOrder)
        if order is not None:
            state = order.get("state")
            if state in [OrderStates.IN_PROGRESS, OrderStates.TOTALED]:
                return True
        return False

    def is_order_in_progress(self, model):
        # type: (eTree) -> bool

        order = model.find(currentOrder)
        if order is not None and order.get("state") == OrderStates.IN_PROGRESS:
            return True
        return False

    def is_order_totaled(
        self,
        model,                  # type: eTree
        order_picture=None,     # type: Optional[eTree]
    ):
        # type: (...) -> bool

        pos_id = model.get("posId")
        order = model.find("CurrentOrder/Order")
        order_id = order.get("orderId")
        pos_ot = self.get_pos_ot(model)

        if order_picture is None:
            order_picture = self.get_order_details(pos_id, order_id, pos_ot)

        order_state = order_picture.find(".//Order").get("state")

        return order is not None and order_state == OrderStates.TOTALED

    def get_current_order(self, model):
        # type: (eTree) -> Optional[eTree]

        return model.find(currentOrder)

    def create_order(self, pos_id, pos_ot, price_list="EI", sale_type="EAT_IN"):
        # type: (int, OrderTaker, Optional[str], Optional[str]) -> int

        order = pos_ot.createOrder(pos_id, price_list, saletype=sale_type)
        return int(order.get("orderId"))

    def get_order_details(self, pos_id, order_id, pos_ot):
        # type: (int, int, OrderTaker) -> eTree

        order_str = pos_ot.orderPicture(pos_id, order_id)
        order = eTree.fromstring(order_str) if order_str != "" else None
        return order

    def add_item_to_order(
        self,
        pos_id,                 # type: int
        pos_ot,                 # type: OrderTaker
        price_list,             # type: str
        item,                   # type: ToAddItem
        model,                  # type: eTree
        sale_type="EAT_IN"      # type: Optional[str]
    ):
        # type: (...) -> None

        context = item.context
        line_number = item.line_number
        product_code = item.product_code
        quantity = item.quantity

        if context and line_number:
            item_id = "{}.{}".format(context, product_code)
            sale_line = self._get_sale_line(
                model=model,
                context=context,
                product_code=product_code,
                line_number=line_number,
            )

            if sale_line is None or str(sale_line.get("itemType")) != "INGREDIENT":
                pos_ot.clearOption(pos_id, line_number, "", item_id)
                pos_ot.doOption(pos_id, item_id, quantity, line_number, '@')
            else:
                pos_ot.changeLineItemQty(pos_id, line_number, quantity, itemid=item_id)
        if context is None and line_number is None:
            item_id = "1.{}".format(product_code)
            pos_ot.doSale(pos_id, item_id, price_list, quantity, verifyOption=False, saletype=sale_type)

    def reopen_shopping_cart(
            self,
            pos_id,     # type: int
            pos_ot      # type: OrderTaker
    ):
        # type: (...) -> eTree

        pos_ot.clearTenders(pos_id)
        pos_ot.reopenOrder(pos_id)
        order_str = pos_ot.orderPicture(int(pos_id))
        order = eTree.fromstring(order_str)
        return order

    def order_has_tender(self, model):
        # type: (eTree) -> Optional[eTree]

        for tender in model.findall("CurrentOrder/Order/TenderHistory/Tender") or []:
            if tender.get('tenderType') != '60':
                return True
        return False

    def void_order(
        self,
        pos_id,         # type: int
        pos_ot,         # type: OrderTaker
        order_id="",    # type: Optional[str]
    ):
        # type: (...) -> None

        pos_ot.voidOrder(
            posid=pos_id,
            orderid=order_id,
        )

    def delete_line(self, pos_id, pos_ot, line_numbers):
        # type: (str, OrderTaker, List[int]) -> None

        line_to_delete = "|".join(map(str, line_numbers))
        pos_ot.voidLine(pos_id, line_to_delete)

    def get_pos_list(self):
        # type: () -> tuple

        return sysactions.get_poslist()

    @staticmethod
    def do_tender(
        pos_id,             # type: int
        tender_id,          # type: int
        pos_ot,             # type: OrderTaker
        amount,             # type: float
        tender_details=""   # type: Optional[str]
    ):
        # type: (...) -> Dict

        return pos_ot.doTender(
            posid=pos_id,
            tenderid=tender_id,
            amount=amount,
            donotclose=True,
            tenderdetail=tender_details
        )

    def set_order_to_paid(self, model, pos_ot):
        # type: (eTree, OrderTaker) -> Any

        order = self.get_current_order(model)
        pos_id = order.get("posId")

        xml_order = get_current_order(model)
        all_tenders = xml_order.findall("TenderHistory/Tender")
        if len(all_tenders) == 0:
            raise Exception("Invalid tender amount")

        last_tender = all_tenders[-1]

        last_amount = last_tender.get("tenderAmount", 0)
        if float(last_amount) == 0:
            raise Exception("Invalid tender amount")

        last_tender_type_id = last_tender.get("tenderType")
        last_tender_id = last_tender.get("tenderId")
        pos_ot.doTender(
            posid=int(pos_id),
            tenderid=last_tender_type_id,
            amount=last_amount,
            donotclose=False,
            ordertenderid=last_tender_id
        )

    def add_custom_property(self, key, value, pos_ot):
        # type: (str, str, OrderTaker) -> None

        pos_ot.setOrderCustomProperty(key, value)

    def add_custom_properties(self,  pos_ot, properties, orderid=''):
        # type: (OrderTaker, dict, str) -> None

        pos_ot.setOrderCustomProperties(properties, orderid)

    def update_order_properties(self, pos_ot, posid, saletype):
        # type: (OrderTaker, str, str) -> eTree

        order = pos_ot.updateOrderProperties(posid, saletype=saletype)
        return order

    def do_total(self, pos_id, pos_ot):
        # type: (int, OrderTaker) -> eTree

        pos_ot.doTotal(pos_id)
        order_str = pos_ot.orderPicture(int(pos_id))
        order = eTree.fromstring(order_str)
        return order

    def get_rupture_products(self):
        # type: () -> list

        message = Message(BusTokens.TK_REMOTE_ORDER_GET_RUPTURED_ITEMS)
        response_message = self.message_bus.send_message("RemoteOrder", message)
        if response_message.token == TK_SYS_ACK:
            return response_message.data

    def get_menu_data(self):
        # type: () -> bytearray

        try:
            with open("menu.json") as menu_file:
                data = self._filter_products(menu_file=menu_file)
                json_data = json.loads(data)
                menu_id = self.get_catalog_version()
                new_json_data = {
                    "products": json_data,
                    "menuId": menu_id if menu_id else ''
                }
                data = json.dumps(new_json_data)

                compress_obj = zlib.compressobj(zlib.Z_BEST_COMPRESSION, zlib.DEFLATED, 31)
                zipped_menu = compress_obj.compress(data)
                zipped_menu += compress_obj.flush()
                return zipped_menu
        except Exception as _:
            self.logger.exception("Error getting menu.json products")

        return ""

    def get_nf_type(self):
        # type: () -> None

        if self.config.nf_type:
            return

        message = Message(BusTokens.TK_FISCALWRAPPER_GET_NF_TYPE)
        response_message = self.message_bus.send_message("FiscalWrapper", message)
        if response_message.token == TK_SYS_ACK:
            self.config.nf_type = response_message.data.upper()
        else:
            self.logger.error(response_message.data)
            raise ExternalServiceErrors(ExternalServicesErrorsType.ServiceUnavailable)

    def get_sefaz_status(self):
        # type: () -> bool

        message = Message(BusTokens.TK_FISCALWRAPPER_SEFAZ_CONNECTIVITY)
        response_message = self.message_bus.send_message("FiscalWrapper", message)
        if response_message.token == TK_SYS_ACK:
            return True if response_message.data else False
        else:
            self.logger.error(response_message.data)

    def get_store_sat_status(
        self,
    ):
        # type: (...) -> bool

        message = Message(token=BusTokens.TK_FISCALWRAPPER_SAT_STORE_STATUS)
        response_message = self.message_bus.send_message(
            component="FiscalWrapper",
            message=message,
        )
        if response_message.token == TK_SYS_ACK:
            result = response_message.data.lower() == "true"
            return result
        else:
            self.logger.error(response_message.data)
            return False

    def get_catalog_version(
        self,
    ):
        # type: (...) -> Optional[str]

        message = Message(BusTokens.TK_POS_UPDATER_GET_VERSIONS, timeout=5000000)
        response_message = self.message_bus.send_message("PosUpdater", message)
        if response_message.token == TK_SYS_ACK:
            data = json.loads(response_message.data or '{}')
            data = data["versions"] if "versions" in data else {}
            return data["catalog"] if "catalog" in data else None
        else:
            self.logger.error(response_message.data)

        return None

    def void_tenders_and_order(self, pos_id, pos_ot):
        # type: (int, OrderTaker) -> Any

        pos_ot.clearTenders(str(pos_id))
        pos_ot.voidOrder(str(pos_id))

    def logout_delivery_user(self, pos_id, operator_id):
        # type: (int, int) -> bool

        message = Message(msgbus.TK_POS_USERLOGOUT,
                          msgbus.FM_PARAM, "%s\0%s" % (pos_id, operator_id),
                          timeout=10000000)
        logout_msg = self.message_bus.send_message("POS%d" % int(pos_id), message)
        if logout_msg.token == msgbus.TK_SYS_ACK:
            return True
        else:
            self.logger.error(logout_msg.data)
            return False

    def close_business_period(self, pos_id):
        # type: (int) -> bool
        self.logger.info("Closing business current period")

        message = Message(msgbus.TK_POS_BUSINESSEND, msgbus.FM_PARAM, "%s" % pos_id, timeout=10000000)
        business_end_msg = self.message_bus.send_message("POS%d" % int(pos_id), message)
        if business_end_msg.token == msgbus.TK_SYS_ACK:
            return True
        else:
            self.logger.error(business_end_msg.data)
            return False

    def open_business_period(self, pos_id):
        # type: (int) -> bool
        self.logger.info("Opening business period")

        current_date = datetime.datetime.now().date().strftime("%Y%m%d")
        message = Message(msgbus.TK_POS_BUSINESSBEGIN,
                          msgbus.FM_PARAM, "%s\0%s" % (int(pos_id), current_date),
                          timeout=10000000)
        business_begin_msg = self.message_bus.send_message("POS%d" % int(pos_id), message)
        if business_begin_msg.token == msgbus.TK_SYS_ACK:
            return True
        else:
            self.logger.error(business_begin_msg.data)
            return False

    def authenticate_user(self, user_id, user_password):
        # type: (str, str) -> bool

        self.logger.info("Authenticate delivery user")
        message = Message(msgbus.TK_USERCTRL_AUTHENTICATE,
                          msgbus.FM_PARAM,
                          '%s\x00%s\x00' % (user_id.encode("utf-8"), user_password.encode("utf-8")),
                          timeout=10000000)
        authenticate_msg = self.message_bus.send_message('UserControl', message)
        if authenticate_msg.token == msgbus.TK_SYS_ACK:
            return authenticate_msg.data
        else:
            self.logger.error(authenticate_msg.data)
            return False

    def open_user(self, pos_id, user_id, long_user_name):
        # type: (int, str, str) -> bool

        self.logger.info("Opening delivery user")
        message = Message(msgbus.TK_POS_USEROPEN,
                          msgbus.FM_PARAM,
                          "%s\0%s\0%s\0%s" % (pos_id, user_id.encode("utf-8"), 0.0, long_user_name),
                          timeout=10000000)
        open_user_message = self.message_bus.send_message("POS%d" % int(pos_id), message)
        if open_user_message.token == msgbus.TK_SYS_ACK:
            return True
        else:
            self.logger.error(open_user_message.data)
            return False

    def login_user(self, pos_id, user_id, long_user_name):

        self.logger.info("Logging delivery user")
        message = Message(msgbus.TK_POS_USERLOGIN,
                          msgbus.FM_PARAM,
                          "%s\0%s\0%s" % (pos_id, user_id.encode("utf-8"), long_user_name),
                          timeout=10000000)
        login_msg = self.message_bus.send_message("POS%d" % int(pos_id), message)
        if login_msg.token == msgbus.TK_SYS_ACK:
            return True
        else:
            self.logger.error(login_msg.data)
            return False

    def get_read_scanner(self, pos_id):
        # type: (int) -> str

        message = "$PUT_CODE_IN_SCANNER_BELOW"
        title = "Scanner"
        timeout = 60000

        model = self.get_model(pos_id)

        current_scanner = get_used_service(model, "scanner")
        if not current_scanner:
            raise ValidationError(PosValidationErrorsType.ScannerNotFound)

        dlg_id = show_messagebox(pos_id, message=message, title=title, buttons="$CANCEL", asynch=True, timeout=timeout,
                                 focus=False)

        try:
            scanner_result = read_scanner(current_scanner, 60000, dlg_id, pos_id)
            if "," in scanner_result:
                response_code, read = scanner_result.split(",")
                if response_code != "0":
                    return
            else:
                return scanner_result
        except Exception as _:
            show_messagebox(pos_id, message="$READING_MISTAKE", icon="error", timeout=180000)

    def print_text(self, pos_id, content):
        # type: (str, str) -> None

        if not content:
            raise ValidationError(PosValidationErrorsType.EmptyPrinterContent)

        model = self.get_model(pos_id)

        current_printer = get_used_service(model, "printer")
        if not current_printer:
            raise ValidationError(PosValidationErrorsType.PrinterNotFound)

        result = print_text(pos_id, model, content, force_printer=current_printer)
        if not result:
            raise ExternalServiceErrors(ExternalServicesErrorsType.ErrorPrinting, current_printer)

    def get_storewide_config(self, parameter, defval=None, array=False):
        # type: (str, Optional[bool], Optional[bool]) -> list

        configs = sysactions.get_storewide_config(parameter, defval, array)

        return configs

    def apply_coupon(
            self,
            pos_id,         # type: int
            coupon_code     # type: str
    ):
        # type: (...) -> (str, bool)

        payload = json.dumps({
            'posId': pos_id,
            'voucherId': coupon_code.upper(),
            'originatorPosId': pos_id
        })

        message = Message(BusTokens.TK_COUPON_APPLY,
                          msgbus.FM_PARAM, payload)
        coupon_msg = self.message_bus.send_message('nCoupon', message)
        if coupon_msg.token == msgbus.TK_SYS_ACK:
            coupon_applied = True
        else:
            self.logger.error(coupon_msg.data)
            coupon_applied = False

        return coupon_msg.data, coupon_applied

    def get_order_custom_properties(self, pos_ot, key, order_id=""):
        # type: (OrderTaker, str, int) -> eTree

        properties = pos_ot.getOrderCustomProperties(key, order_id)
        return properties

    def get_cfg(self, pos_id, key, default):
        # type: (int, str, str) -> Any

        configs = sysactions.get_cfg(pos_id).key_value(key, default)
        return configs

    def change_line_quantity(self, model, pos_ot, request):
        # type: (eTree.Element, OrderTaker, List[ToChangeItems]) -> Any

        pos_id = model.get("posId")
        to_change_items = []

        for item in request:
            if item.new_quantity == 0:
                raise ValidationError(PosValidationErrorsType.InvalidQuantity)

            order_sale_lines = get_current_order(model).findall("SaleLine")
            selected_sale_line = None
            for sale_line in order_sale_lines:
                current_line_number = sale_line.get("lineNumber")
                current_level = sale_line.get("level")
                current_code = int(sale_line.get("partCode") or 0)
                current_context = sale_line.get("itemId")
                if current_context == item.context and current_code == item.product_code and \
                        current_line_number == str(item.line_number) and current_level == "0":
                    selected_sale_line = sale_line
                    break

            if selected_sale_line is None:
                raise ValidationError(PosValidationErrorsType.SaleItemsInvalid)

            item_type = selected_sale_line.get("itemType")

            if item_type == "COMBO":
                raise ValidationError(PosValidationErrorsType.ItemTypeComboInvalid)

            to_change_items.append(item)

        for item in to_change_items:
            item_id = "{}.{}".format(item.context, item.product_code)
            pos_ot.changeLineItemQty(pos_id, str(item.line_number), item.new_quantity, item_id)

    def remove_coupon(
            self,
            pos_id,             # type: int
            coupon_applied_id   # type: str
    ):
        # type: (...) -> (str, bool)

        data = json.dumps({
            'pos_id': pos_id,
            'benefit': coupon_applied_id
        })

        message = Message(
            token=BusTokens.TK_COUPON_REMOVE,
            data_type=msgbus.FM_PARAM,
            data=data
        )
        response = self.message_bus.send_message('nDiscount', message)

        if response.token == msgbus.TK_SYS_ACK:
            coupon_removed = True
        else:
            self.logger.error(response.data)
            coupon_removed = False

        return response.data, coupon_removed

    def _get_languages(
        self,
        uppercase=False     # type: Optional[bool]
    ):
        # type: (...) -> list

        message = Message(
            token=BusTokens.TK_I18N_GETLANGS,
            data_type=msgbus.FM_PARAM
        )
        response = self.message_bus.send_message(
            component="I18N",
            message=message
        )
        languages = []
        if response.token == TK_SYS_ACK:
            languages = sorted(set(re.findall(r'[a-z]{2}', response.data)))
        if uppercase:
            languages = [l.upper() for l in languages]

        return languages

    def get_text_from_i18n(
            self
    ):
        # type: () -> Dict

        l10n_params = [
            "L10_PYTHON_DATE_FORMAT",
            "L10_PYTHON_DATETIME_FORMAT",
            "L10N_DECIMAL_PLACES",
            "L10N_CURRENCY_DECIMALS",
            "L10N_CURRENCY_SYMBOL",
            "L10N_DECIMALS_SEPARATOR",
            "L10N_THOUSANDS_SEPARATOR",
            "L10N_CURRENCY_SYMBOL_POSITION"
        ]
        languages = self._get_languages()

        text = {}
        for language in languages:
            l10n_param = language + "\0"
            l10n_param += "\0".join(l10n_params)

            message = Message(
                token=BusTokens.TK_I18N_GETTABLEJSON,
                data_type=msgbus.FM_PARAM,
                data=l10n_param
            )
            response_message = self.message_bus.send_message(
                component="I18N",
                message=message
            )
            if response_message.token == TK_SYS_ACK:
                text[self.languages_translations[language.upper()]] = json.loads(response_message.data)
            else:
                raise ExternalServiceErrors(ExternalServicesErrorsType.LanguageValuesNotFound, language.upper())

        return text

    def get_tender_types(
            self
    ):
        # type: () -> Any

        return sysactions.get_tender_types()

    @staticmethod
    def _get_customer_info(
            data    # type: Dict[str, Any]
    ):
        # type: (...) -> Dict[str, str]

        customer_info = data.get("customerInfo", "")
        document_info = customer_info.get("document", None)
        name_info = customer_info.get("name", None)
        valid_states = ["onStart", "onTotal"]

        document = None
        if document_info and any([d in valid_states for d in document_info]):
            document = "INITIAL" if "onStart" in document_info else "FINAL"

        name = None
        if name_info and any([n in valid_states for n in name_info]):
            name = "INITIAL" if "onStart" in name_info else "FINAL"

        return dict(document=document, name=name)

    def _send_message_to_pos(
            self,
            data,   # type: str
            pos_id  # type: int
    ):
        # type: (...) -> Message

        message = Message(
            token=BusTokens.TK_POS_EXECUTEACTION,
            data_type=msgbus.FM_PARAM,
            data=data
        )
        return self.message_bus.send_message(
            component="POS{}".format(pos_id),
            message=message
        )

    def get_totem_languages(
        self
    ):
        # type: (...) -> List[str]

        all_languages = self._get_languages(uppercase=True)
        if not all_languages:
            return []

        if not self.config.totem_config_path or not os.path.exists(self.config.totem_config_path):
            return []

        config_file = open(self.config.totem_config_path, 'rb')
        data = json.load(config_file)
        totem_languages = data.get("languages", [])

        filtered_languages = [self.languages_translations[l] for l in all_languages if l in totem_languages]

        return filtered_languages

    def get_static_config(
        self,
        pos_id  # type: str
    ):
        # type: (...) -> Dict[str, Any]

        data = "\0".join(map(str, ["get_static_config", pos_id]))
        response = self._send_message_to_pos(
            data=data,
            pos_id=pos_id
        )

        if response.token != msgbus.TK_SYS_ACK:
            raise ExternalServiceErrors(ExternalServicesErrorsType.CannotGetStoreConfigForPos, pos_id)

        data = json.loads(response.data or "{}")
        if data:
            customer_info = self._get_customer_info(data)

            available_sale_types = data.get("availableSaleTypes", {})
            sale_type_quick = available_sale_types.get("saleTypeQuick", {})
            show_take_out_screen = any(["TAKE_OUT" in sale_types for sale_types in sale_type_quick])

            totem_configuration = data.get("totemConfigurations", {})
            totem_products = totem_configuration.get("products", {})
            languages = self.get_totem_languages()
            cancel_timeout_window = data.get("cancelTimeoutWindow", -1)
            screen_timeout = data.get("screenTimeout", -1)
            discount_config = totem_configuration.get("discount", {})
            show_discount = discount_config.get("showDiscount", True)

            return dict(
                CUSTOMER_INFO=customer_info,
                SHOW_TAKE_OUT_SCREEN=show_take_out_screen,
                SHOW_SCANNER_SCREEN=data.get("totemQRCodeEnabled", False),
                SHOW_RUPTURE_ITEMS=data.get("showRupturedProducts", False),
                SHOW_POPUP_COMBO=totem_products.get("showPopupCombo", False),
                LANGUAGES=languages,
                CANCEL_TIMEOUT_WINDOW=cancel_timeout_window,
                SCREEN_TIMEOUT=screen_timeout,
                SHOW_DISCOUNTS=show_discount,
            )

        raise ValidationError(PosValidationErrorsType.EmptyStoreConfigurationForPos)

    def do_clear_tender(
            self,
            pos_id,      # type: str
            model,       # type: object
            tender_id    # type: int
    ):
        # type: (...) -> float

        pos_ot = self.get_pos_ot(model)

        current_order = sysactions.get_current_order(model)
        tenders = current_order.findall(".//TenderHistory/Tender")

        tender_ids = [int(t.get("tenderId")) for t in tenders if t.get("tenderId", None) is not None]

        if tender_id not in tender_ids:
            raise ValidationError(PosValidationErrorsType.InvalidTenderId)

        pos_ot.clearTenders(
            pos_id,
            None,
            tender_id
        )

        current_order = pos_ot.orderPicture(pos_id)
        current_order = eTree.XML(current_order)
        return current_order.get("dueAmount")

    def get_navigation(
            self,
            pos_id  # type: int
    ):
        # type: (...) -> Dict

        data = "\0".join(map(str, ["get_navigation_data", pos_id]))
        response = self._send_message_to_pos(
            data=data,
            pos_id=pos_id
        )

        if response.token != msgbus.TK_SYS_ACK:
            raise ExternalServiceErrors(ExternalServicesErrorsType.CannotGetNavigationForPos, pos_id)

        navigation = json.loads(response.data or "{}")

        if not self.descriptions:
            self.descriptions = self.navigation_service.get_navigation_descriptions()

        self._get_navigation_descriptions(
            parent_nav=navigation,
        )

        return navigation

    def _get_navigation_descriptions(
        self,
        parent_nav,    # type: dict
    ):
        # type: (...) -> None

        for nav in parent_nav:
            nav_id = nav.get("nav_id")
            nav["navigation_text"] = {
                "pt-br": "",
                "en-us": "",
                "es-es": "",
            }

            if nav_id in self.descriptions:
                nav["navigation_text"] = self.descriptions.get(nav_id)

            groups = nav.get("groups")
            if groups:
                self._get_navigation_descriptions(
                    parent_nav=groups,
                )

    def do_stored_order(
        self,
        pos_id,     # type: int
        model,      # type: eTree
    ):
        # type: (...) -> None

        pos_ot = self.get_pos_ot(
           model=model,
        )

        pos_ot.storeOrder(
           posid=pos_id,
        )

    def get_manager_status(
        self,
        pos_id,     # type: int
    ):
        # type: (...) -> bool

        model = self.get_model(pos_id=pos_id)
        totem_was_closed = sysactions.get_custom(
            model=model,
            key="TOTEM_WAS_CLOSED",
        )
        if totem_was_closed is None:
            return False
        else:
            if upper(totem_was_closed) == "TRUE":
                return True
            elif upper(totem_was_closed) == "FALSE":
                return False
            else:
                raise Exception("Totem status unavailable")

    def _filter_products(
        self,
        menu_file,  # type: file
    ):
        # type: (...) -> str

        products = json.load(fp=menu_file)
        filtered_products = self._filter_sons(
            products=products,
            pos_type="TT",
        )

        return json.dumps(obj=filtered_products)

    def _filter_sons(
        self,
        products,   # type: List[Dict]
        pos_type,   # type: str
    ):
        # type: (...) -> List[Dict]

        filtered_products = []
        for product in products:
            if pos_type not in product.get("product", {}).get("hideOnPosType", []):
                product["sons"] = self._filter_sons(
                    products=product.get("sons", []),
                    pos_type=pos_type,
                )
                filtered_products.append(product)

        return filtered_products

    def _get_sale_line(
        self,
        model,              # type: eTree
        context,            # type: str
        product_code,       # type: str
        line_number,        # type: str
    ):
        # type: (...) -> Optional[eTree.Element]

        sale_lines = get_current_order(model=model).findall("SaleLine")
        for line in sale_lines:
            exists = self._exists_item(
                line=line,
                context=context,
                product_code=product_code,
                line_number=line_number,
            )

            if exists:
                return line

        return None

    @staticmethod
    def _exists_item(
        line,           # type: eTree.Element
        context,        # type: str
        product_code,   # type: str
        line_number,    # type: str
    ):
        sale_line_line_number = str(line.get("lineNumber"))
        sale_line_code = str(line.get("partCode") or "")
        sale_line_context = str(line.get("itemId"))

        return (
            sale_line_context == str(context) and
            sale_line_code == str(product_code) and
            sale_line_line_number == str(line_number)
        )




import sys
import traceback

from ._DataType import DataType
from ._DefaultToken import DefaultToken
from ._Message import Message


class InternalErrorMessage(Message):
    def __init__(
        self,
        ex,    # type: Exception
    ):
        tb = sys.exc_info()[2]
        exception_messages = traceback.format_exception(type(ex), ex, tb)
        super(InternalErrorMessage, self).__init__(
            token=DefaultToken.TK_SYS_NAK,
            data_type=DataType.string,
            data="".join(exception_messages).encode("utf-8"),
        )



import sys

if sys.version_info[0] < 3:
    import inspect2 as inspect
else:
    import inspect

import json
from typing import Dict, Union, Tuple

from messagebus import HttpResponseMessage
from ._util import convert_to_json


class ExceptionFormatter(object):
    def __init__(self, exception_map):
        # type: (Dict[type, Union[int, Tuple[int, str]]]) -> None
        self.exception_map = exception_map

    def format(self, message_bus, message, exception, default_status_code):
        try:
            map_info = self.exception_map[type(exception)]
            if isinstance(map_info, int):
                exception_cls = type(exception)
                body = b""
                headers = {}
                if self.has_own_init(exception_cls):
                    body = convert_to_json(exception).encode("utf-8")
                    headers["Content-Type"] = "application/json; charset=UTF-8"

                message_bus.reply_message(message, HttpResponseMessage(map_info, body, headers))
            else:
                status_code = map_info[0]
                body = map_info[1]
                if isinstance(body, dict):
                    message_bus.reply_message(message,
                                              HttpResponseMessage(
                                                  status_code,
                                                  json.dumps(body).encode("utf-8"),
                                                  {"Content-Type": "application/json; charset=UTF-8"}))
                else:
                    message_bus.reply_message(message, HttpResponseMessage(status_code, body))

        except TypeError:
            self._create_response_when_exception_map_is_none(message_bus, message, default_status_code)
        except KeyError:
            self._create_response_when_exception_map_does_not_have_exception(message_bus,
                                                                             message,
                                                                             exception,
                                                                             default_status_code)

    @staticmethod
    def _create_response_when_exception_map_is_none(message_bus, message, default_status_code):
        message_bus.reply_message(message, HttpResponseMessage(default_status_code))

    @staticmethod
    def _create_response_when_exception_map_does_not_have_exception(message_bus,
                                                                    message,
                                                                    exception,
                                                                    default_status_code):
        cls = type(exception)
        status_code = default_status_code
        body = b""

        if hasattr(cls, "_cls_attrs"):
            status_code = cls._cls_attrs.get("status_code", status_code)
            body = cls._cls_attrs.get("body", b"")

        if isinstance(body, dict):
            message_bus.reply_message(message,
                                      HttpResponseMessage(
                                          status_code,
                                          json.dumps(body).encode("utf-8"),
                                          {"Content-Type": "application/json; charset=UTF-8"}))
        else:
            message_bus.reply_message(message, HttpResponseMessage(status_code, body))

    @staticmethod
    def has_own_init(cls):
        mro = inspect.getmro(cls)
        for i in range(1, len(mro)):
            if mro[i].__init__ == cls.__init__:
                return False
        return True



from typing import List
from xml.etree import ElementTree
from pos_model import TaxItem
from xml.sax.saxutils import escape


class SaleLineUtil(object):
    def __init__(self):
        return

    def get_product_name_from_tax_item(self, tax_item, sale_lines):
        # type: (TaxItem, List[ElementTree]) -> unicode
        sale_line = self._get_sale_line_by_tax_item(tax_item, sale_lines)

        if sale_line is None:
            raise Exception("SaleLine not found for product: {0}.{1}".format(tax_item.item_id, tax_item.part_code))

        product_name = escape(sale_line.get("productName"))

        return product_name

    def get_sale_line_by_tax_item(self, tax_item, sale_lines):
        # type: (TaxItem, List[ElementTree]) -> Optional[ElementTree]
        for sale_line in sale_lines:
            if sale_line.get("itemId") == tax_item.item_id and int(sale_line.get("partCode")) == tax_item.part_code:
                return sale_line

        raise Exception("SaleLine not found for product: {0}.{1}".format(tax_item.item_id, tax_item.part_code))



import json

import sysactions

from bustoken import (
    TK_RUPTURA_GET_DAILY_DISHES,
    TK_RUPTURA_GET_DISABLED,
    TK_RUPTURA_GET_ENABLED
)
from model.customexception import GetRuptureItemsException
from msgbus import TK_SYS_NAK, FM_STRING
from typing import Dict


def get_disabled_products():
    msg = sysactions.send_message("Ruptura", TK_RUPTURA_GET_DISABLED, FM_STRING, "UI")

    if msg.token == TK_SYS_NAK:
        raise GetRuptureItemsException("Error getting rupture disable products")

    return json.loads(msg.data)


def get_enabled_products():
    msg = sysactions.send_message("Ruptura", TK_RUPTURA_GET_ENABLED, FM_STRING, "UI")
    if msg.token == TK_SYS_NAK:
        raise GetRuptureItemsException("Error getting rupture enabled products")

    return json.loads(msg.data)


def get_daily_dishes():
    # type: (...) -> Dict

    msg = sysactions.send_message(
        dest_name="Ruptura",
        token=TK_RUPTURA_GET_DAILY_DISHES,
        format=FM_STRING
    )

    if msg.token == TK_SYS_NAK:
        raise GetRuptureItemsException("Error getting rupture daily dishes")

    return json.loads(msg.data)



from application.model import Configurations
from application.repository import NDiscountRepository
from typing import List


class GetBenefitsInteractor(object):
    def __init__(self, configs, repository):
        # type: (Configurations, NDiscountRepository) -> None
        self.configs = configs
        self.logger = configs.logger
        self.repository = repository

    def apply(self, only_visible):
        # type: (bool) -> List[dict]

        return self.repository.get_benefits_by_visibility_from_database(only_visible)



from helper import BaseRepository
from persistence import Connection  # noqa
from typing import List, Tuple  # noqa

from reports_app.cashreport.dto import Order  # noqa


class FiscalRepository(BaseRepository):
    def __init__(self, mbcontext):
        super(FiscalRepository, self).__init__(mbcontext)

    def get_card_brands(self, paid_orders):
        # type: (List[Order]) -> List[Tuple[unicode, int, float, int]]
        def inner_func(conn):
            # type: (Connection) -> List[Tuple[unicode, int, float, int]]
            order_list = map(lambda order: order.order_id, paid_orders)
            order_list = ','.join(map(str, order_list))
            query = self._CardBrandsQuery
            query = query.format(order_list)
            return [((x.get_entry(0)).decode('utf-8'), int(x.get_entry(1)), float(x.get_entry(2)), int(x.get_entry(3))) for x in conn.select(query)]

        return self.execute_with_connection(inner_func, None, "FiscalPersistence")

    _CardBrandsQuery = """select
                          case when bc.Descricao is not null then bc.Descricao else 'OUTROS' end as Brand, pd.Type as Type, sum(pd.Amount) as Amount, count(pd.OrderId)
                          from fiscal.PaymentData pd
                          left join fiscal.BandeiraCartao bc on pd.Bandeira = bc.Bandeira
                          where pd.Type in (1,2) and pd.OrderId in ({0})
                          group by bc.Descricao, pd.Type
                          order by Type asc, Amount desc;"""



from ._OrderCondition import OrderCondition
from ._ProdStateCondition import ProdStateCondition
from ._ConditionBox import ConditionBox
from ._CustomPropertyCondition import CustomPropertyCondition
from ._PodTypeCondition import PodTypeCondition
from ._CustomPropertyValuesCondition import CustomPropertyValuesCondition
from ._OrderStateCondition import OrderStateCondition
from ._DisplayTimeCondition import DisplayTimeCondition
from ._SaleTypeCondition import SaleTypeCondition




from ._TipReportHeaderPresenter import TipReportHeaderPresenter
from ._TipReportBodyPresenter import TipReportBodyPresenter
__all__ = [
    "TipReportHeaderPresenter",
    "TipReportBodyPresenter"
]



from typing import Dict,  Optional


class HttpResponse(object):
    def __init__(self, status_code=200, body=None, headers=None):
        # type: (int, Optional[bytes], Optional[Dict[str, str]]) -> None
        self.status_code = status_code
        self.body = body
        self.headers = headers

    def __repr__(self):
        return 'HttpResponse(status_code = {}, body = {}, header = "{}")'.format(self.status_code,
                                                                                 self.body,
                                                                                 self.headers)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, o):
        if not isinstance(o, HttpResponse):
            return False

        return self.status_code == o.status_code and self.headers == o.headers and self.body == o.body

    def __ne__(self, o):
        return not self.__eq__(o)



import json

from xml.etree import cElementTree as eTree

from messagebus import DefaultToken, DataType, Message
from messageprocessorutil import UuidMessageProcessor

from application.interactor import OnSaleItemInteractor
from application.model import Configurations


class OnSaleItemProcessor(UuidMessageProcessor):
    def __init__(self, configs, interactor):
        # type: (Configurations, OnSaleItemInteractor) -> None
        super(OnSaleItemProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (str, str) -> (int, int, int, str, eTree)

        data = json.loads(data)
        pos_id = int(data.get("pos_id"))

        return pos_id

    def call_business(self, pos_id):
        # type: (int) -> None

        return self.interactor.execute(pos_id=pos_id)

    def format_response(self, message_bus, message, event, request, response):
        if message:
            data = eTree.tostring(element=response) if response is not None else ""
            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=DefaultToken.TK_SYS_ACK,
                    data_type=DataType.string,
                    data=data,
                ),
            )

    def format_exception(self, message_bus, message, event, input_obj, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from ._ApiErrorsType import PosValidationErrorsType
from ._ApiErrorsType import AuthErrorsType
from ._ApiErrorsType import OperationErrorsType
from ._ApiErrorsType import ExternalServicesErrorsType
from ._ApiErrorsType import BadRequestErrorsType
from ._ApiErrorsType import NotFoundErrorsType



import json
from time import sleep

from application.model import OperatorCashInOut, TransferOrderInfo
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType
from msgbus import TK_STORECFG_GET, TK_SYS_NAK, TK_POS_GETPOSLIST
from mwhelper import BaseRepository
from persistence import Connection
from typing import Optional, Dict, List

FISCAL_WRAPPER_GET_XML_TOKEN = 51380248
TK_SLCTRL_OMGR_ORDERPICT = 0x00A00002


class MessageBusRepository(BaseRepository):

    def __init__(self, message_bus):
        # type: (MbContextMessageBus) -> None
        super(MessageBusRepository, self).__init__(message_bus.mbcontext)

        self.message_bus = message_bus
        self.pos_list = None

    def read_sw_config(self, key):
        # type: (str) -> Optional[str]

        message = Message(TK_STORECFG_GET, data_type=DataType.param, data=key)
        msg = self.message_bus.send_message("StoreWideConfig", message)
        if msg.token == TK_SYS_NAK:
            return None
        return str(msg.data)

    def get_fiscal_xml_from_fiscal_wrapper(self, order_id):
        # type: (int) -> Optional[str]

        message = Message(FISCAL_WRAPPER_GET_XML_TOKEN, data_type=DataType.param, data=order_id)
        msg = self.message_bus.send_message("FiscalWrapper", message)
        if msg.token == TK_SYS_NAK:
            return None

        return str(msg.data) if msg.data else None

    def get_operator_time(self, session_id):
        # type: (str) -> Dict

        def inner_func(conn):
            # type: (Connection) -> Dict

            query = """SELECT strftime('%Y-%m-%dT%H:%M:%f', OpenTime) AS OpenTime,
                              strftime('%Y-%m-%dT%H:%M:%f', CloseTime) AS CloseTime
                       FROM UserSession
                       WHERE SessionId = '{}'""".format(session_id)
            cursor = conn.select(query)

            operator_time = {}
            for row in cursor:
                operator_time["OpenTime"] = row.get_entry("OpenTime")
                operator_time["CloseTime"] = row.get_entry("CloseTime")

            return operator_time

        return self.execute_with_connection(inner_func, service_name="Persistence")

    def get_transfers_by_session_id(self, pos_id, session_id):
        # type: (str, str) -> Dict

        def inner_func(conn):
            # type: (Connection) -> List

            query = """SELECT TenderId, Type, Amount FROM Transfer WHERE SessionId='{}'"""
            cursor = conn.select(query.format(session_id))

            database_transfers = {}
            for row in cursor:
                tender_id = row.get_entry("TenderId")
                if tender_id not in database_transfers:
                    database_transfers[tender_id] = {}

                transfer_type = row.get_entry("Type")
                transfer_amount = float(row.get_entry("Amount") or 0)
                if transfer_type not in database_transfers[tender_id]:
                    database_transfers[tender_id][transfer_type] = 0

                database_transfers[tender_id][transfer_type] += transfer_amount

            return database_transfers

        return self.execute_with_transaction(inner_func, str(pos_id))

    def get_transfer_order_info(self, session_id):
        transfer_order_info = {}

        def inner_func(conn):
            # type: (Connection) -> List

            query = """
                        SELECT 
                            ot.TenderId, 
                            COUNT(ot.TenderId) AS TenderQuantity, 
                            SUM(ot.TipAmount) AS TipAmount, 
                            SUM(ot.TenderAmount) - SUM(COALESCE(ot.ChangeAmount,0)) AS TotalTender
                        FROM Orders o
                        JOIN OrderTender ot ON ot.OrderId = o.OrderId
                        WHERE SessionId='{}'
                        GROUP BY ot.tenderId
                    """
            cursor = conn.select(query.format(session_id))

            for row in cursor:
                tender_id = row.get_entry("TenderId")
                if tender_id not in transfer_order_info:
                    transfer_order_info[tender_id] = TransferOrderInfo()

                transfer_order_info[tender_id].payment_quantity += int(row.get_entry("TenderQuantity") or 0)
                transfer_order_info[tender_id].tip_amount += float(row.get_entry("TipAmount") or 0)
                transfer_order_info[tender_id].total_tender += float(row.get_entry("TotalTender") or 0)

            return transfer_order_info

        self.execute_in_all_databases(inner_func, self._get_pos_list())
        return transfer_order_info

    def get_cash_in_out_by_session_id(self, session_id):
        # type: (str) -> List

        def inner_func(conn):
            # type: (Connection) -> List

            query = "SELECT * FROM Transfer WHERE SessionId = '{0}' AND Description IN ('{1}', '{2}')"\
                .format(session_id, OperatorCashInOut.CASH_IN, OperatorCashInOut.CASH_OUT)
            cursor = conn.select(query)

            cash_in_out_list = []
            for row in cursor:
                transfer = {"description": row.get_entry("Description"),
                            "manager_id": _get_manager_id(row),
                            "tender_type": row.get_entry("TenderId"),
                            "amount": float(row.get_entry("Amount") or 0),
                            "timestamp": row.get_entry("Timestamp")}
                cash_in_out_list.append(transfer)

            return cash_in_out_list

        def _get_manager_id(row):
            try:
                manager_id = row.get_entry("ManagerID")
            except ValueError:
                gl_account = row.get_entry("GLAccount")
                try:
                    parsed_gla_account = json.loads(gl_account)
                    manager_id = parsed_gla_account.get("manager", None)
                except (ValueError, TypeError):
                    manager_id = None

            return int(manager_id) if manager_id else None

        return self.execute_with_connection(inner_func)

    def get_product_categories(self, full=False):
        # type: (Optional[bool]) -> List

        def inner_func(conn):
            # type: (Connection) -> List

            if full:
                query = "SELECT * FROM ProductCustomParams"
            else:
                query = """SELECT ProductCode, CustomParamId, CustomParamValue
                           FROM (
                               SELECT
                                   ProductCode,
                                   substr(Tag,1,instr(Tag,'=')-1) AS CustomParamId,
                                   substr(Tag,instr(Tag,'=')+1) AS CustomParamValue
                               FROM ProductTags
                           ) Tags
                           WHERE Tags.CustomParamId IN ('Categories', 'SubCategories')"""
            cursor = conn.select(query)

            product_categories = []
            for row in cursor:
                product = {"ProductCode": row.get_entry("ProductCode"),
                           "CustomParamId": row.get_entry("CustomParamId"),
                           "CustomParamValue": row.get_entry("CustomParamValue")}
                product_categories.append(product)

            return product_categories

        return self.execute_with_connection(inner_func, service_name="Persistence")

    def get_declared_amount(self, business_period):
        # type: (str) -> Dict

        def inner_func(conn):
            # type: (Connection) -> Dict

            query = """SELECT TenderId, tdsum(Amount) 
                       FROM Transfer 
                       WHERE SessionId='{}' AND Type=6 
                       GROUP BY TenderId""".format(business_period)
            cursor = conn.select(query)

            declared_amounts = dict([(int(row.get_entry(0)), float(row.get_entry(1) or 0)) for row in cursor])
            return declared_amounts

        return self.execute_with_connection(inner_func, service_name="Persistence")

    def _get_last_order_id(self):
        # type: () -> int

        def inner_func(conn):
            # type: (Connection) -> int

            query = """SELECT MAX(SeqNo) FROM Sequencer WHERE SeqNm='OrderIdGen'"""
            cursor = conn.select(query)

            max_order_id = None
            for row in cursor:
                max_order_id = int(row.get_entry(0))
            return max_order_id

        return self.execute_with_connection(inner_func, service_name="Persistence")

    def get_last_order_picture(self):
        order_id = self._get_last_order_id()
        if order_id == 0:
            return None
        return self.get_order_picture(order_id)

    def get_order_picture(self, order_id):
        pos_id = ""
        params = (pos_id, str(order_id), "true")
        req_buf = "\0".join(str(p) for p in params)

        message = Message(TK_SLCTRL_OMGR_ORDERPICT, data_type=DataType.param, data=req_buf)
        msg = self.message_bus.send_message("ORDERMGR0", message)

        if msg.token == TK_SYS_NAK:
            return None
        params = msg.data.split('\0')
        return params[2]

    def _get_pos_list(self):
        # type: () -> List

        if self.pos_list is not None:
            return self.pos_list

        retry_cnt = 10
        while retry_cnt:
            msg = self.mb_context.MB_EasySendMessage("PosController", TK_POS_GETPOSLIST)
            if msg.token == TK_SYS_NAK:
                retry_cnt -= 1
                sleep(1)
            else:
                self.pos_list = sorted(map(int, msg.data.split("\0")))
                retry_cnt = 0

        return self.pos_list



from typing import Dict


class Action(object):
    def __init__(
        self,
        obj,     # type: Dict
    ):
        # type: (...) -> None

        self.order = obj.get("order", None)
        self.name = obj.get("name")
        self.value = obj.get("value")
        self.times_to_apply = None



# -*- coding: utf-8 -*-
import json

from msgbus import MBEasyContext
from pos_model import OrderParser, SaleItem
from xml.etree import cElementTree as eTree

from _GeneralTaxCalculator import GeneralTaxCalculator
from _OrderFormatter import OrderFormatter
from taxcalculator import ReducedBaseAmountTaxCalculator
from typing import Dict


class TaxCalculatorService(object):
    def __init__(self, order_parser, general_tax_calculator, order_formatter, mb_context):
        # type: (OrderParser, GeneralTaxCalculator, OrderFormatter, MBEasyContext) -> None

        self.order_parser = order_parser
        self.general_tax_calculator = general_tax_calculator
        self.order_formatter = order_formatter
        self.mb_context = mb_context
        self.calculators = {}

    def calculate_taxes(self, order_str):
        # type: (str) -> unicode
        xml = eTree.XML(order_str)
        root = xml if xml.tag == "Order" else None
        order_xml = root or xml.find(".//Order")
        if not order_xml:
            raise Exception("Invalid order XML")

        order = self.order_parser.parse_order(order_xml)
        for sale_item in order.sale_items:
            tax_items = self.general_tax_calculator.calculate_all_taxes(sale_item)
            sale_item.taxes = tax_items

        formatted_order = self.order_formatter.format_order(order)

        return formatted_order

    def insert_tx_idx(self, order_str):
        # type: (str) -> unicode
        order_xml = eTree.XML(order_str).find("Order")
        for sale_line in order_xml.findall("SaleLine"):
            if float(sale_line.get("qty") or 1) == 0:
                continue

            sale_item = self.order_parser.create_sale_item(sale_line)
            tax_items = self.general_tax_calculator.calculate_all_taxes(sale_item)
            sale_line_price = sale_item.unit_price if sale_item.unit_price > 0 else sale_item.added_unit_price
            if tax_items and float(sale_line_price) > 0:
                sale_item.taxes = tax_items
                tax_item = sale_item.get_tax_item("ICMS")
                if tax_item is None:
                    continue

                sale_line.set("taxIndex", tax_item.tax_index)
        return eTree.tostring(order_xml)

    def calculate_taxes_item(self, sale_line):
        # type: (str) -> unicode

        sale_line = json.loads(sale_line)

        if float(sale_line.get("qty") or 1) == 0:
            return []
        part_code = str(sale_line.get("partCode"))
        sale_item = SaleItem(
                part_code=part_code,
                product_name=sale_line.get("productName"),
                quantity=sale_line.get("quantity"),
                unit_price=sale_line.get("unitPrice"),
                item_price=sale_line.get("itemPrice")
        )

        calculators = self._get_taxes_calculator(sale_line, part_code)
        taxes_items = self.general_tax_calculator.calculate_all_taxes(sale_item, calculators)

        return json.dumps(taxes_items, default=lambda x: x.__dict__)

    def _get_taxes_calculator(self, sale_line, part_code):
        # type: (Dict, str) -> []

        if part_code in self.calculators.keys():
            return self.calculators[part_code]

        taxes_calculate = json.loads(sale_line.get("taxesCalc"))
        base_reduction_dict = self._get_sale_line_tax_dictionary(sale_line, "baseReduction")
        cst_dict = self._get_sale_line_tax_dictionary(sale_line, "cst")
        try:
            data = []
            for tax in taxes_calculate:
                index = str(tax.get("fiscalIndex"))
                code = tax.get("code")
                name = str(tax.get("name"))
                products = dict([(str(key), str(value)) for key, value in tax.get("product").items()])
                rate = float(tax.get("rate"))
                params = 'ICMS' if name in ['PIS', 'COFINS'] else None
                calculator = ReducedBaseAmountTaxCalculator(
                    self.mb_context,
                    code,
                    name,
                    index,
                    rate,
                    products,
                    params,
                    cst_dict,
                    base_reduction_dict
                )
                data.append(calculator)
            self.calculators[part_code] = data
            return data
        except Exception as _:
            return {}

    @staticmethod
    def _get_sale_line_tax_dictionary(
        sale_line,
        tax_name
    ):
        # type: (...) -> Dict

        response = {}
        sale_line_tax = sale_line.get(tax_name)
        if sale_line_tax is not None:
            response = sale_line_tax.items()
            response = dict([(str(key), str(value)) for key, value in response])

        return response




from sysactions import action, get_model, print_report


@action
def show_pos_versions(pos_id, pos_versions_json):
    model = get_model(pos_id)
    print_report(pos_id, model, True, "versions_report", pos_id, pos_versions_json)



class XmlEnveloper(object):
    NAMESPACE_SOAP = "http://www.w3.org/2003/05/soap-envelope"

    def __init__(self, c_uf):
        self.c_uf = str(c_uf)

    def envelop(self, xml, webservice_namespace, versao_ws):
        prefix = '<Envelope xmlns="{0}">'\
                     '<Header>'\
                         '<nfeCabecMsg xmlns="{1}">'\
                             '<cUF>{2}</cUF>'\
                             '<versaoDados>{3:.2f}</versaoDados>'\
                             '</nfeCabecMsg>'\
                      '</Header>'\
                      '<Body>'\
                          '<nfeDadosMsg xmlns="{1}">'\
            .format(XmlEnveloper.NAMESPACE_SOAP, webservice_namespace, self.c_uf, 3.1 if versao_ws in (1, 3) else 4)
        suffix = "</nfeDadosMsg></Body></Envelope>"
        return prefix + xml + suffix



from application.interactor import LoyaltyAvailableInteractor
from application.model import AsyncUuidMessageProcessor
from messagebus import DefaultToken, DataType, Message


class LoyaltyAvailableProcessor(AsyncUuidMessageProcessor):
    def __init__(self, interactor):
        # type: (LoyaltyAvailableInteractor) -> None

        super(LoyaltyAvailableProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, _):
        # it does not expect any data
        pass

    def call_business(self, _):
        # type: (None) -> bool

        return self.interactor.is_loyalty_available()

    def format_response(self, message_bus, message, event, request, response):
        if response:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK))
        else:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))
